# Copyright (c) 2021 AccelByte Inc. All Rights Reserved.
# This is licensed software from AccelByte Inc, for limitations
# and restrictions contact your company contract manager.
#
# Code generated. DO NOT EDIT!

# template file: ags_py_codegen

# AccelByte Gaming Services Seasonpass Service

# pylint: disable=duplicate-code
# pylint: disable=line-too-long
# pylint: disable=missing-function-docstring
# pylint: disable=missing-module-docstring
# pylint: disable=too-many-arguments
# pylint: disable=too-many-branches
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-lines
# pylint: disable=too-many-locals
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-statements
# pylint: disable=unused-import

from .utils import randomize

from ..api.seasonpass.models import BulkUserProgressionRequest
from ..api.seasonpass.models import ClaimableRewards
from ..api.seasonpass.models import ClaimableUserSeasonInfo
from ..api.seasonpass.models import ErrorEntity
from ..api.seasonpass.models import ExcessStrategy
from ..api.seasonpass.models import ExpGrantHistoryInfo
from ..api.seasonpass.models import ExpGrantHistoryPagingSlicedResult
from ..api.seasonpass.models import FieldValidationError
from ..api.seasonpass.models import FullSeasonInfo
from ..api.seasonpass.models import Image
from ..api.seasonpass.models import ListSeasonInfo
from ..api.seasonpass.models import ListSeasonInfoPagingSlicedResult
from ..api.seasonpass.models import ListUserSeasonInfo
from ..api.seasonpass.models import ListUserSeasonInfoPagingSlicedResult
from ..api.seasonpass.models import Localization
from ..api.seasonpass.models import LocalizedPassInfo
from ..api.seasonpass.models import LocalizedSeasonInfo
from ..api.seasonpass.models import Ownership
from ..api.seasonpass.models import Paging
from ..api.seasonpass.models import PassCreate
from ..api.seasonpass.models import PassInfo
from ..api.seasonpass.models import PassUpdate
from ..api.seasonpass.models import Permission
from ..api.seasonpass.models import ReasonTagsResult
from ..api.seasonpass.models import RewardCreate
from ..api.seasonpass.models import RewardCurrency
from ..api.seasonpass.models import RewardInfo
from ..api.seasonpass.models import RewardUpdate
from ..api.seasonpass.models import SeasonCloneRequest
from ..api.seasonpass.models import SeasonCreate
from ..api.seasonpass.models import SeasonInfo
from ..api.seasonpass.models import SeasonSummary
from ..api.seasonpass.models import SeasonUpdate
from ..api.seasonpass.models import Tier
from ..api.seasonpass.models import TierCreate
from ..api.seasonpass.models import TierInput
from ..api.seasonpass.models import TierPagingSlicedResult
from ..api.seasonpass.models import TierReorder
from ..api.seasonpass.models import UserExpGrant
from ..api.seasonpass.models import UserPassGrant
from ..api.seasonpass.models import UserPurchasable
from ..api.seasonpass.models import UserRewardClaim
from ..api.seasonpass.models import UserSeasonSummary
from ..api.seasonpass.models import UserTierGrant
from ..api.seasonpass.models import ValidationErrorEntity


def create_bulk_user_progression_request_example() -> BulkUserProgressionRequest:
    instance = BulkUserProgressionRequest()
    instance.user_ids = [randomize()]
    return instance


def create_claimable_rewards_example() -> ClaimableRewards:
    instance = ClaimableRewards()
    instance.claiming_rewards = {randomize(): randomize()}
    instance.to_claim_rewards = {randomize(): randomize()}
    return instance


def create_claimable_user_season_info_example() -> ClaimableUserSeasonInfo:
    instance = ClaimableUserSeasonInfo()
    instance.claiming_rewards = {randomize(): randomize()}
    instance.cleared = randomize("bool")
    instance.created_at = randomize("date")
    instance.current_exp = randomize("int", min_val=1, max_val=1000)
    instance.current_tier_index = randomize("int", min_val=1, max_val=1000)
    instance.enrolled_at = randomize("date")
    instance.enrolled_passes = [randomize()]
    instance.id_ = randomize()
    instance.last_tier_index = randomize("int", min_val=1, max_val=1000)
    instance.namespace = randomize("slug")
    instance.required_exp = randomize("int", min_val=1, max_val=1000)
    instance.season_id = randomize()
    instance.to_claim_rewards = {randomize(): randomize()}
    instance.updated_at = randomize("date")
    instance.user_id = randomize("uid")
    instance.season = create_season_summary_example()
    instance.total_exp = randomize("int", min_val=1, max_val=1000)
    instance.total_paid_for_exp = randomize("int", min_val=1, max_val=1000)
    instance.total_sweat_exp = randomize("int", min_val=1, max_val=1000)
    return instance


def create_error_entity_example() -> ErrorEntity:
    instance = ErrorEntity()
    instance.error_code = randomize("int", min_val=1, max_val=1000)
    instance.error_message = randomize()
    instance.dev_stack_trace = randomize()
    instance.message_variables = {randomize(): randomize()}
    instance.required_permission = create_permission_example()
    return instance


def create_excess_strategy_example() -> ExcessStrategy:
    instance = ExcessStrategy()
    instance.method = randomize()
    instance.currency = randomize()
    instance.percent_per_exp = randomize("int", min_val=1, max_val=1000)
    return instance


def create_exp_grant_history_info_example() -> ExpGrantHistoryInfo:
    instance = ExpGrantHistoryInfo()
    instance.created_at = randomize("date")
    instance.grant_exp = randomize("int", min_val=1, max_val=1000)
    instance.id_ = randomize()
    instance.namespace = randomize("slug")
    instance.season_id = randomize()
    instance.user_id = randomize("uid")
    instance.source = randomize()
    instance.tags = [randomize()]
    return instance


def create_exp_grant_history_paging_sliced_result_example() -> (
    ExpGrantHistoryPagingSlicedResult
):
    instance = ExpGrantHistoryPagingSlicedResult()
    instance.data = [create_exp_grant_history_info_example()]
    instance.paging = create_paging_example()
    instance.total = randomize("int", min_val=1, max_val=1000)
    return instance


def create_field_validation_error_example() -> FieldValidationError:
    instance = FieldValidationError()
    instance.error_code = randomize()
    instance.error_field = randomize()
    instance.error_message = randomize()
    instance.error_value = randomize()
    instance.message_variables = {randomize(): randomize()}
    return instance


def create_full_season_info_example() -> FullSeasonInfo:
    instance = FullSeasonInfo()
    instance.id_ = randomize()
    instance.localizations = {}
    instance.name = randomize()
    instance.namespace = randomize("slug")
    instance.passes = [create_pass_info_example()]
    instance.rewards = {}
    instance.tiers = [create_tier_example()]
    instance.images = [create_image_example()]
    return instance


def create_image_example() -> Image:
    instance = Image()
    instance.height = randomize("int", min_val=1, max_val=1000)
    instance.image_url = randomize("url")
    instance.width = randomize("int", min_val=1, max_val=1000)
    instance.as_ = randomize()
    instance.caption = randomize()
    instance.small_image_url = randomize("url")
    return instance


def create_list_season_info_example() -> ListSeasonInfo:
    instance = ListSeasonInfo()
    instance.created_at = randomize("date")
    instance.default_language = randomize()
    instance.end = randomize("date")
    instance.id_ = randomize()
    instance.name = randomize()
    instance.namespace = randomize("slug")
    instance.start = randomize("date")
    instance.status = randomize()
    instance.tier_item_id = randomize()
    instance.tier_item_name = randomize()
    instance.updated_at = randomize("date")
    instance.pass_codes = [randomize()]
    instance.published_at = randomize("date")
    return instance


def create_list_season_info_paging_sliced_result_example() -> (
    ListSeasonInfoPagingSlicedResult
):
    instance = ListSeasonInfoPagingSlicedResult()
    instance.data = [create_list_season_info_example()]
    instance.paging = create_paging_example()
    return instance


def create_list_user_season_info_example() -> ListUserSeasonInfo:
    instance = ListUserSeasonInfo()
    instance.cleared = randomize("bool")
    instance.current_tier_index = randomize("int", min_val=1, max_val=1000)
    instance.enrolled_at = randomize("date")
    instance.id_ = randomize()
    instance.last_tier_index = randomize("int", min_val=1, max_val=1000)
    instance.namespace = randomize("slug")
    instance.season_id = randomize()
    instance.user_id = randomize("uid")
    instance.season = create_season_summary_example()
    return instance


def create_list_user_season_info_paging_sliced_result_example() -> (
    ListUserSeasonInfoPagingSlicedResult
):
    instance = ListUserSeasonInfoPagingSlicedResult()
    instance.data = [create_list_user_season_info_example()]
    instance.paging = create_paging_example()
    instance.total = randomize("int", min_val=1, max_val=1000)
    return instance


def create_localization_example() -> Localization:
    instance = Localization()
    instance.description = randomize()
    instance.title = randomize()
    return instance


def create_localized_pass_info_example() -> LocalizedPassInfo:
    instance = LocalizedPassInfo()
    instance.auto_enroll = randomize("bool")
    instance.code = randomize()
    instance.created_at = randomize("date")
    instance.display_order = randomize()
    instance.language = randomize()
    instance.namespace = randomize("slug")
    instance.pass_item_id = randomize()
    instance.season_id = randomize()
    instance.updated_at = randomize("date")
    instance.description = randomize()
    instance.images = [create_image_example()]
    instance.title = randomize()
    return instance


def create_localized_season_info_example() -> LocalizedSeasonInfo:
    instance = LocalizedSeasonInfo()
    instance.auto_claim = randomize("bool")
    instance.created_at = randomize("date")
    instance.end = randomize("date")
    instance.id_ = randomize()
    instance.language = randomize()
    instance.name = randomize()
    instance.namespace = randomize("slug")
    instance.passes = [create_localized_pass_info_example()]
    instance.rewards = {}
    instance.start = randomize("date")
    instance.status = randomize()
    instance.tier_item_id = randomize()
    instance.tiers = [create_tier_example()]
    instance.updated_at = randomize("date")
    instance.description = randomize()
    instance.images = [create_image_example()]
    instance.pass_codes = [randomize()]
    instance.published_at = randomize("date")
    instance.title = randomize()
    return instance


def create_ownership_example() -> Ownership:
    instance = Ownership()
    instance.owned = randomize("bool")
    return instance


def create_paging_example() -> Paging:
    instance = Paging()
    instance.next_ = randomize()
    instance.previous = randomize()
    return instance


def create_pass_create_example() -> PassCreate:
    instance = PassCreate()
    instance.code = randomize()
    instance.display_order = randomize("int", min_val=1, max_val=1000)
    instance.localizations = {}
    instance.pass_item_id = randomize()
    instance.auto_enroll = randomize("bool")
    instance.images = [create_image_example()]
    return instance


def create_pass_info_example() -> PassInfo:
    instance = PassInfo()
    instance.auto_enroll = randomize("bool")
    instance.code = randomize()
    instance.created_at = randomize("date")
    instance.display_order = randomize()
    instance.localizations = {}
    instance.namespace = randomize("slug")
    instance.pass_item_id = randomize()
    instance.pass_item_name = randomize()
    instance.season_id = randomize()
    instance.updated_at = randomize("date")
    instance.images = [create_image_example()]
    return instance


def create_pass_update_example() -> PassUpdate:
    instance = PassUpdate()
    instance.auto_enroll = randomize("bool")
    instance.display_order = randomize("int", min_val=1, max_val=1000)
    instance.images = [create_image_example()]
    instance.localizations = {}
    instance.pass_item_id = randomize()
    return instance


def create_permission_example() -> Permission:
    instance = Permission()
    instance.action = randomize("int", min_val=1, max_val=1000)
    instance.resource = randomize()
    return instance


def create_reason_tags_result_example() -> ReasonTagsResult:
    instance = ReasonTagsResult()
    instance.tags = [randomize()]
    return instance


def create_reward_create_example() -> RewardCreate:
    instance = RewardCreate()
    instance.code = randomize()
    instance.type_ = randomize()
    instance.currency = create_reward_currency_example()
    instance.image = create_image_example()
    instance.item_id = randomize()
    instance.quantity = randomize("int", min_val=1, max_val=1000)
    return instance


def create_reward_currency_example() -> RewardCurrency:
    instance = RewardCurrency()
    instance.currency_code = randomize()
    instance.namespace = randomize("slug")
    return instance


def create_reward_info_example() -> RewardInfo:
    instance = RewardInfo()
    instance.code = randomize()
    instance.namespace = randomize("slug")
    instance.season_id = randomize()
    instance.type_ = randomize()
    instance.currency = create_reward_currency_example()
    instance.image = create_image_example()
    instance.item_id = randomize()
    instance.item_name = randomize()
    instance.item_sku = randomize()
    instance.item_type = randomize()
    instance.quantity = randomize("int", min_val=1, max_val=1000)
    return instance


def create_reward_update_example() -> RewardUpdate:
    instance = RewardUpdate()
    instance.currency = create_reward_currency_example()
    instance.image = create_image_example()
    instance.item_id = randomize()
    instance.null_fields = [randomize()]
    instance.quantity = randomize("int", min_val=1, max_val=1000)
    instance.type_ = randomize()
    return instance


def create_season_clone_request_example() -> SeasonCloneRequest:
    instance = SeasonCloneRequest()
    instance.end = randomize("date")
    instance.name = randomize()
    instance.start = randomize("date")
    return instance


def create_season_create_example() -> SeasonCreate:
    instance = SeasonCreate()
    instance.default_required_exp = randomize("int", min_val=1, max_val=1000)
    instance.draft_store_id = randomize()
    instance.end = randomize("date")
    instance.localizations = {}
    instance.name = randomize()
    instance.start = randomize("date")
    instance.tier_item_id = randomize()
    instance.auto_claim = randomize("bool")
    instance.default_language = randomize()
    instance.excess_strategy = create_excess_strategy_example()
    instance.images = [create_image_example()]
    return instance


def create_season_info_example() -> SeasonInfo:
    instance = SeasonInfo()
    instance.auto_claim = randomize("bool")
    instance.created_at = randomize("date")
    instance.default_language = randomize()
    instance.default_required_exp = randomize("int", min_val=1, max_val=1000)
    instance.draft_store_id = randomize()
    instance.end = randomize("date")
    instance.excess_strategy = create_excess_strategy_example()
    instance.id_ = randomize()
    instance.localizations = {}
    instance.name = randomize()
    instance.namespace = randomize("slug")
    instance.start = randomize("date")
    instance.status = randomize()
    instance.tier_item_id = randomize()
    instance.tier_item_name = randomize()
    instance.updated_at = randomize("date")
    instance.images = [create_image_example()]
    instance.pass_codes = [randomize()]
    instance.published_at = randomize("date")
    return instance


def create_season_summary_example() -> SeasonSummary:
    instance = SeasonSummary()
    instance.end = randomize("date")
    instance.id_ = randomize()
    instance.name = randomize()
    instance.namespace = randomize("slug")
    instance.start = randomize("date")
    instance.status = randomize()
    instance.pass_codes = [randomize()]
    instance.previous = create_season_summary_example()
    instance.published_at = randomize("date")
    return instance


def create_season_update_example() -> SeasonUpdate:
    instance = SeasonUpdate()
    instance.auto_claim = randomize("bool")
    instance.default_language = randomize()
    instance.default_required_exp = randomize("int", min_val=1, max_val=1000)
    instance.draft_store_id = randomize()
    instance.end = randomize("date")
    instance.excess_strategy = create_excess_strategy_example()
    instance.images = [create_image_example()]
    instance.localizations = {}
    instance.name = randomize()
    instance.start = randomize("date")
    instance.tier_item_id = randomize()
    return instance


def create_tier_example() -> Tier:
    instance = Tier()
    instance.id_ = randomize()
    instance.required_exp = randomize("int", min_val=1, max_val=1000)
    instance.rewards = {}
    return instance


def create_tier_create_example() -> TierCreate:
    instance = TierCreate()
    instance.index = randomize("int", min_val=1, max_val=1000)
    instance.quantity = randomize("int", min_val=1, max_val=1000)
    instance.tier = create_tier_input_example()
    return instance


def create_tier_input_example() -> TierInput:
    instance = TierInput()
    instance.required_exp = randomize("int", min_val=1, max_val=1000)
    instance.rewards = {}
    return instance


def create_tier_paging_sliced_result_example() -> TierPagingSlicedResult:
    instance = TierPagingSlicedResult()
    instance.data = [create_tier_example()]
    instance.paging = create_paging_example()
    instance.total = randomize("int", min_val=1, max_val=1000)
    return instance


def create_tier_reorder_example() -> TierReorder:
    instance = TierReorder()
    instance.new_index = randomize("int", min_val=1, max_val=1000)
    return instance


def create_user_exp_grant_example() -> UserExpGrant:
    instance = UserExpGrant()
    instance.exp = randomize("int", min_val=1, max_val=1000)
    instance.source = randomize()
    instance.tags = [randomize()]
    return instance


def create_user_pass_grant_example() -> UserPassGrant:
    instance = UserPassGrant()
    instance.pass_code = randomize()
    instance.pass_item_id = randomize()
    return instance


def create_user_purchasable_example() -> UserPurchasable:
    instance = UserPurchasable()
    instance.pass_item_id = randomize()
    instance.tier_item_count = randomize("int", min_val=1, max_val=1000)
    instance.tier_item_id = randomize()
    return instance


def create_user_reward_claim_example() -> UserRewardClaim:
    instance = UserRewardClaim()
    instance.pass_code = randomize()
    instance.reward_code = randomize()
    instance.tier_index = randomize("int", min_val=1, max_val=1000)
    return instance


def create_user_season_summary_example() -> UserSeasonSummary:
    instance = UserSeasonSummary()
    instance.namespace = randomize("slug")
    instance.user_id = randomize("uid")
    instance.cleared = randomize("bool")
    instance.created_at = randomize("date")
    instance.current_exp = randomize("int", min_val=1, max_val=1000)
    instance.current_tier_index = randomize("int", min_val=1, max_val=1000)
    instance.enrolled_at = randomize("date")
    instance.enrolled_passes = [randomize()]
    instance.id_ = randomize()
    instance.last_tier_index = randomize("int", min_val=1, max_val=1000)
    instance.required_exp = randomize("int", min_val=1, max_val=1000)
    instance.season_id = randomize()
    instance.updated_at = randomize("date")
    return instance


def create_user_tier_grant_example() -> UserTierGrant:
    instance = UserTierGrant()
    instance.count = randomize("int", min_val=1, max_val=1000)
    instance.source = randomize()
    instance.tags = [randomize()]
    return instance


def create_validation_error_entity_example() -> ValidationErrorEntity:
    instance = ValidationErrorEntity()
    instance.error_code = randomize("int", min_val=1, max_val=1000)
    instance.error_message = randomize()
    instance.errors = [create_field_validation_error_example()]
    return instance



# Copyright (c) 2021 AccelByte Inc. All Rights Reserved.
# This is licensed software from AccelByte Inc, for limitations
# and restrictions contact your company contract manager.
#
# Code generated. DO NOT EDIT!

# template file: ags_py_codegen

"""Auto-generated package that contains models used by the AccelByte Gaming Services Challenge Service."""

__version__ = "1.9.0"
__author__ = "AccelByte"
__email__ = "dev@accelbyte.net"

# pylint: disable=line-too-long

from .iam_error_response import IamErrorResponse
from .iam_permission import IamPermission
from .model_challenge_response import ModelChallengeResponse
from .model_challenge_response import (
    AssignmentRuleEnum as ModelChallengeResponseAssignmentRuleEnum,
    GoalsVisibilityEnum as ModelChallengeResponseGoalsVisibilityEnum,
    RotationEnum as ModelChallengeResponseRotationEnum,
    StatusEnum as ModelChallengeResponseStatusEnum,
)
from .model_claim_user_rewards_req import ModelClaimUserRewardsReq
from .model_claim_users_rewards_request import ModelClaimUsersRewardsRequest
from .model_claim_users_rewards_response import ModelClaimUsersRewardsResponse
from .model_create_challenge_request import ModelCreateChallengeRequest
from .model_create_challenge_request import (
    AssignmentRuleEnum as ModelCreateChallengeRequestAssignmentRuleEnum,
    GoalsVisibilityEnum as ModelCreateChallengeRequestGoalsVisibilityEnum,
    RotationEnum as ModelCreateChallengeRequestRotationEnum,
)
from .model_create_goal_request import ModelCreateGoalRequest
from .model_evaluate_player_progression_request import (
    ModelEvaluatePlayerProgressionRequest,
)
from .model_get_goals_response import ModelGetGoalsResponse
from .model_goal_meta import ModelGoalMeta
from .model_goal_order import ModelGoalOrder
from .model_goal_progression_response import ModelGoalProgressionResponse
from .model_goal_progression_response import (
    StatusEnum as ModelGoalProgressionResponseStatusEnum,
)
from .model_goal_response import ModelGoalResponse
from .model_goal_schedule import ModelGoalSchedule
from .model_list_challenge_response import ModelListChallengeResponse
from .model_list_periods_response import ModelListPeriodsResponse
from .model_list_user_rewards_response import ModelListUserRewardsResponse
from .model_pagination import ModelPagination
from .model_plugin_asignment_app_config import ModelPluginAsignmentAppConfig
from .model_plugin_assignment_custom_config import ModelPluginAssignmentCustomConfig
from .model_plugin_assignment_request import ModelPluginAssignmentRequest
from .model_plugin_assignment_request import (
    ExtendTypeEnum as ModelPluginAssignmentRequestExtendTypeEnum,
)
from .model_plugin_assignment_response import ModelPluginAssignmentResponse
from .model_plugin_assignment_response import (
    ExtendTypeEnum as ModelPluginAssignmentResponseExtendTypeEnum,
)
from .model_predicate import ModelPredicate
from .model_predicate import (
    MatcherEnum as ModelPredicateMatcherEnum,
    ParameterTypeEnum as ModelPredicateParameterTypeEnum,
)
from .model_requirement import ModelRequirement
from .model_requirement import OperatorEnum as ModelRequirementOperatorEnum
from .model_requirement_progression_response import ModelRequirementProgressionResponse
from .model_reset_config import ModelResetConfig
from .model_reward import ModelReward
from .model_reward import TypeEnum as ModelRewardTypeEnum
from .model_schedule import ModelSchedule
from .model_update_challenge_request import ModelUpdateChallengeRequest
from .model_update_challenge_request import (
    AssignmentRuleEnum as ModelUpdateChallengeRequestAssignmentRuleEnum,
    GoalsVisibilityEnum as ModelUpdateChallengeRequestGoalsVisibilityEnum,
    RotationEnum as ModelUpdateChallengeRequestRotationEnum,
)
from .model_update_goal_request import ModelUpdateGoalRequest
from .model_user_progression_period_response_meta import (
    ModelUserProgressionPeriodResponseMeta,
)
from .model_user_progression_response import ModelUserProgressionResponse
from .model_user_progression_response_meta import ModelUserProgressionResponseMeta
from .model_user_reward import ModelUserReward
from .model_user_reward import (
    StatusEnum as ModelUserRewardStatusEnum,
    TypeEnum as ModelUserRewardTypeEnum,
)
from .models_period import ModelsPeriod
from .response_error import ResponseError



# Copyright (c) 2021 AccelByte Inc. All Rights Reserved.
# This is licensed software from AccelByte Inc, for limitations
# and restrictions contact your company contract manager.
#
# Code generated. DO NOT EDIT!

# template file: ags_py_codegen

"""Auto-generated package that contains models used by the AccelByte Gaming Services Session Service."""

__version__ = "3.17.1"
__author__ = "AccelByte"
__email__ = "dev@accelbyte.net"

# pylint: disable=line-too-long

from .admin_delete_bulk_game__144132 import AdminDeleteBulkGameSessions
from .admin_kick_game_session_member import AdminKickGameSessionMember
from .admin_query_game_sessions import AdminQueryGameSessions
from .admin_query_game_sessio_9fd853 import AdminQueryGameSessionsByAttributes
from .admin_set_ds_ready import AdminSetDSReady
from .admin_update_game_sessi_718745 import AdminUpdateGameSessionMember
from .append_team_game_session import AppendTeamGameSession
from .create_game_session import CreateGameSession
from .delete_game_session import DeleteGameSession
from .game_session_generate_code import GameSessionGenerateCode
from .get_game_session import GetGameSession
from .get_game_session_by_pod_name import GetGameSessionByPodName
from .get_session_server_secret import GetSessionServerSecret
from .join_game_session import JoinGameSession
from .leave_game_session import LeaveGameSession
from .patch_update_game_session import PatchUpdateGameSession
from .public_game_session_cancel import PublicGameSessionCancel
from .public_game_session_invite import PublicGameSessionInvite
from .public_game_session_reject import PublicGameSessionReject
from .public_kick_game_sessio_89864f import PublicKickGameSessionMember
from .public_promote_game_ses_482ff3 import PublicPromoteGameSessionLeader
from .public_query_game_sessi_875927 import PublicQueryGameSessionsByAttributes
from .public_query_my_game_sessions import PublicQueryMyGameSessions
from .public_revoke_game_sess_c2a52e import PublicRevokeGameSessionCode
from .public_session_join_code import PublicSessionJoinCode
from .update_game_session import UpdateGameSession
from .update_game_session_bac_df8b84 import UpdateGameSessionBackfillTicketID



# Copyright (c) 2021 AccelByte Inc. All Rights Reserved.
# This is licensed software from AccelByte Inc, for limitations
# and restrictions contact your company contract manager.
#
# Code generated. DO NOT EDIT!

# template file: ags_py_codegen

# pylint: disable=duplicate-code
# pylint: disable=line-too-long
# pylint: disable=missing-function-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=missing-module-docstring
# pylint: disable=too-many-arguments
# pylint: disable=too-many-branches
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-lines
# pylint: disable=too-many-locals
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-statements
# pylint: disable=unused-import

from typing import Any, Dict, List, Optional, Tuple, Union

from ....core import HeaderStr
from ....core import get_namespace as get_services_namespace
from ....core import run_request
from ....core import run_request_async
from ....core import deprecated
from ....core import same_doc_as

from ..models import ModelsEventRegistry

from ..operations.event_registry import GetRegisteredEventIDHandler
from ..operations.event_registry import GetRegisteredEventsByEventTypeHandler
from ..operations.event_registry import GetRegisteredEventsHandler
from ..operations.event_registry import RegisterEventHandler
from ..operations.event_registry import UnregisterEventIDHandler
from ..operations.event_registry import UpdateEventRegistryHandler


@deprecated
@same_doc_as(GetRegisteredEventIDHandler)
def get_registered_event_id_handler(
    event_id: str, x_additional_headers: Optional[Dict[str, str]] = None, **kwargs
):
    """Get registered eventID data from the Event Registry (GetRegisteredEventIDHandler)

    Required permission `ADMIN:NAMESPACE:{namespace}:EVENT [READ]`and scope `analytics`

    Properties:
        url: /event/registry/eventIds/{eventId}

        method: GET

        tags: ["Event Registry"]

        consumes: []

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        event_id: (eventId) REQUIRED str in path

    Responses:
        200: OK - ModelsEventRegistry (OK)

        400: Bad Request - (Bad Request)

        401: Unauthorized - (Unauthorized)

        403: Forbidden - (Forbidden)

        404: Not Found - (Not Found)

        500: Internal Server Error - (Internal Server Error)
    """
    request = GetRegisteredEventIDHandler.create(
        event_id=event_id,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@deprecated
@same_doc_as(GetRegisteredEventIDHandler)
async def get_registered_event_id_handler_async(
    event_id: str, x_additional_headers: Optional[Dict[str, str]] = None, **kwargs
):
    """Get registered eventID data from the Event Registry (GetRegisteredEventIDHandler)

    Required permission `ADMIN:NAMESPACE:{namespace}:EVENT [READ]`and scope `analytics`

    Properties:
        url: /event/registry/eventIds/{eventId}

        method: GET

        tags: ["Event Registry"]

        consumes: []

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        event_id: (eventId) REQUIRED str in path

    Responses:
        200: OK - ModelsEventRegistry (OK)

        400: Bad Request - (Bad Request)

        401: Unauthorized - (Unauthorized)

        403: Forbidden - (Forbidden)

        404: Not Found - (Not Found)

        500: Internal Server Error - (Internal Server Error)
    """
    request = GetRegisteredEventIDHandler.create(
        event_id=event_id,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@deprecated
@same_doc_as(GetRegisteredEventsByEventTypeHandler)
def get_registered_events_by_event_type_handler(
    event_type: str, x_additional_headers: Optional[Dict[str, str]] = None, **kwargs
):
    """Get registered eventID data from the Event Registry based on the event type (GetRegisteredEventsByEventTypeHandler)

    Required permission `ADMIN:NAMESPACE:{namespace}:EVENT [READ]`and scope `analytics`

    Properties:
        url: /event/registry/eventTypes/{eventType}

        method: GET

        tags: ["Event Registry"]

        consumes: []

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        event_type: (eventType) REQUIRED str in path

    Responses:
        200: OK - ModelsEventRegistry (OK)

        400: Bad Request - (Bad Request)

        401: Unauthorized - (Unauthorized)

        403: Forbidden - (Forbidden)

        404: Not Found - (Not Found)

        500: Internal Server Error - (Internal Server Error)
    """
    request = GetRegisteredEventsByEventTypeHandler.create(
        event_type=event_type,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@deprecated
@same_doc_as(GetRegisteredEventsByEventTypeHandler)
async def get_registered_events_by_event_type_handler_async(
    event_type: str, x_additional_headers: Optional[Dict[str, str]] = None, **kwargs
):
    """Get registered eventID data from the Event Registry based on the event type (GetRegisteredEventsByEventTypeHandler)

    Required permission `ADMIN:NAMESPACE:{namespace}:EVENT [READ]`and scope `analytics`

    Properties:
        url: /event/registry/eventTypes/{eventType}

        method: GET

        tags: ["Event Registry"]

        consumes: []

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        event_type: (eventType) REQUIRED str in path

    Responses:
        200: OK - ModelsEventRegistry (OK)

        400: Bad Request - (Bad Request)

        401: Unauthorized - (Unauthorized)

        403: Forbidden - (Forbidden)

        404: Not Found - (Not Found)

        500: Internal Server Error - (Internal Server Error)
    """
    request = GetRegisteredEventsByEventTypeHandler.create(
        event_type=event_type,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@deprecated
@same_doc_as(GetRegisteredEventsHandler)
def get_registered_events_handler(
    x_additional_headers: Optional[Dict[str, str]] = None, **kwargs
):
    """Get registered events in the Event Registry (GetRegisteredEventsHandler)

    Required permission `ADMIN:NAMESPACE:{namespace}:EVENT [READ]`and scope `analytics`

    Properties:
        url: /event/registry/eventIds

        method: GET

        tags: ["Event Registry"]

        consumes: []

        produces: ["application/json"]

        securities: [BEARER_AUTH]

    Responses:
        200: OK - ModelsEventRegistry (OK)

        401: Unauthorized - (Unauthorized)

        403: Forbidden - (Forbidden)

        500: Internal Server Error - (Internal Server Error)
    """
    request = GetRegisteredEventsHandler.create()
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@deprecated
@same_doc_as(GetRegisteredEventsHandler)
async def get_registered_events_handler_async(
    x_additional_headers: Optional[Dict[str, str]] = None, **kwargs
):
    """Get registered events in the Event Registry (GetRegisteredEventsHandler)

    Required permission `ADMIN:NAMESPACE:{namespace}:EVENT [READ]`and scope `analytics`

    Properties:
        url: /event/registry/eventIds

        method: GET

        tags: ["Event Registry"]

        consumes: []

        produces: ["application/json"]

        securities: [BEARER_AUTH]

    Responses:
        200: OK - ModelsEventRegistry (OK)

        401: Unauthorized - (Unauthorized)

        403: Forbidden - (Forbidden)

        500: Internal Server Error - (Internal Server Error)
    """
    request = GetRegisteredEventsHandler.create()
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@deprecated
@same_doc_as(RegisterEventHandler)
def register_event_handler(
    body: ModelsEventRegistry,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Register an eventID to the Event Registry (RegisterEventHandler)

    Required permission `ADMIN:NAMESPACE:{namespace}:EVENT [CREATE]`and scope `analytics`

    Properties:
        url: /event/registry/eventIds

        method: POST

        tags: ["Event Registry"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        body: (body) REQUIRED ModelsEventRegistry in body

    Responses:
        201: Created - (Created)

        400: Bad Request - (Bad Request)

        401: Unauthorized - (Unauthorized)

        403: Forbidden - (Forbidden)

        409: Conflict - (Conflict)

        500: Internal Server Error - (Internal Server Error)
    """
    request = RegisterEventHandler.create(
        body=body,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@deprecated
@same_doc_as(RegisterEventHandler)
async def register_event_handler_async(
    body: ModelsEventRegistry,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Register an eventID to the Event Registry (RegisterEventHandler)

    Required permission `ADMIN:NAMESPACE:{namespace}:EVENT [CREATE]`and scope `analytics`

    Properties:
        url: /event/registry/eventIds

        method: POST

        tags: ["Event Registry"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        body: (body) REQUIRED ModelsEventRegistry in body

    Responses:
        201: Created - (Created)

        400: Bad Request - (Bad Request)

        401: Unauthorized - (Unauthorized)

        403: Forbidden - (Forbidden)

        409: Conflict - (Conflict)

        500: Internal Server Error - (Internal Server Error)
    """
    request = RegisterEventHandler.create(
        body=body,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@deprecated
@same_doc_as(UnregisterEventIDHandler)
def unregister_event_id_handler(
    event_id: str, x_additional_headers: Optional[Dict[str, str]] = None, **kwargs
):
    """Unregister eventID from the Event Registry (UnregisterEventIDHandler)

    Required permission `ADMIN:NAMESPACE:{namespace}:EVENT [DELETE]`and scope `analytics`

    Properties:
        url: /event/registry/eventIds/{eventId}

        method: DELETE

        tags: ["Event Registry"]

        consumes: []

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        event_id: (eventId) REQUIRED str in path

    Responses:
        204: No Content - (No Content)

        401: Unauthorized - (Unauthorized)

        403: Forbidden - (Forbidden)

        404: Not Found - (Not Found)

        500: Internal Server Error - (Internal Server Error)
    """
    request = UnregisterEventIDHandler.create(
        event_id=event_id,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@deprecated
@same_doc_as(UnregisterEventIDHandler)
async def unregister_event_id_handler_async(
    event_id: str, x_additional_headers: Optional[Dict[str, str]] = None, **kwargs
):
    """Unregister eventID from the Event Registry (UnregisterEventIDHandler)

    Required permission `ADMIN:NAMESPACE:{namespace}:EVENT [DELETE]`and scope `analytics`

    Properties:
        url: /event/registry/eventIds/{eventId}

        method: DELETE

        tags: ["Event Registry"]

        consumes: []

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        event_id: (eventId) REQUIRED str in path

    Responses:
        204: No Content - (No Content)

        401: Unauthorized - (Unauthorized)

        403: Forbidden - (Forbidden)

        404: Not Found - (Not Found)

        500: Internal Server Error - (Internal Server Error)
    """
    request = UnregisterEventIDHandler.create(
        event_id=event_id,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@deprecated
@same_doc_as(UpdateEventRegistryHandler)
def update_event_registry_handler(
    body: ModelsEventRegistry,
    event_id: str,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Update an eventID data in the Event Registry (UpdateEventRegistryHandler)

    Required permission `ADMIN:NAMESPACE:{namespace}:EVENT [UPDATE]`and scope `analytics`

    Properties:
        url: /event/registry/eventIds/{eventId}

        method: POST

        tags: ["Event Registry"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        body: (body) REQUIRED ModelsEventRegistry in body

        event_id: (eventId) REQUIRED str in path

    Responses:
        204: No Content - (No Content)

        400: Bad Request - (Bad Request)

        401: Unauthorized - (Unauthorized)

        403: Forbidden - (Forbidden)

        404: Not Found - (Not Found)

        500: Internal Server Error - (Internal Server Error)
    """
    request = UpdateEventRegistryHandler.create(
        body=body,
        event_id=event_id,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@deprecated
@same_doc_as(UpdateEventRegistryHandler)
async def update_event_registry_handler_async(
    body: ModelsEventRegistry,
    event_id: str,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Update an eventID data in the Event Registry (UpdateEventRegistryHandler)

    Required permission `ADMIN:NAMESPACE:{namespace}:EVENT [UPDATE]`and scope `analytics`

    Properties:
        url: /event/registry/eventIds/{eventId}

        method: POST

        tags: ["Event Registry"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        body: (body) REQUIRED ModelsEventRegistry in body

        event_id: (eventId) REQUIRED str in path

    Responses:
        204: No Content - (No Content)

        400: Bad Request - (Bad Request)

        401: Unauthorized - (Unauthorized)

        403: Forbidden - (Forbidden)

        404: Not Found - (Not Found)

        500: Internal Server Error - (Internal Server Error)
    """
    request = UpdateEventRegistryHandler.create(
        body=body,
        event_id=event_id,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )



# Copyright (c) 2024 AccelByte Inc. All Rights Reserved.
# This is licensed software from AccelByte Inc, for limitations
# and restrictions contact your company contract manager.

from __future__ import annotations

import asyncio
import math
from copy import deepcopy
from enum import Enum, auto
from typing import Any, Dict, List, Optional, Union
from typing import Protocol, runtime_checkable

import websockets
import websockets.exceptions

from ._utils import create_basic_authentication


def empty_queue(queue: asyncio.Queue) -> None:
    while not queue.empty():
        queue.get_nowait()
        queue.task_done()


def make_url(url: str, **kwargs) -> str:
    scheme, base = url.split("://", maxsplit=1)

    if scheme.lower() not in ("ws", "wss"):
        scheme = kwargs.get("scheme", None)
        if scheme is None:
            scheme = kwargs.get("uri_prefix", None)  # backwards compatibility
        if scheme is None:
            scheme = "wss"

    scheme = scheme.removesuffix("://")

    suffix = kwargs.get("url_suffix", "")
    if not suffix:
        suffix = kwargs.get("uri_suffix", "")  # backwards compatibility

    return f"{scheme}://{base}{suffix}"


def make_headers(**kwargs) -> Optional[Dict[str, str]]:
    headers = {}

    if (
        (extra_headers := kwargs.get("extra_headers", {}))
        and isinstance(extra_headers, dict)
        and extra_headers
    ):
        headers.update(extra_headers)

    if (access_token := kwargs.get("access_token", None)) and isinstance(
        access_token, str
    ):
        headers["Authorization"] = f"Bearer {access_token}"
    elif (username := kwargs.get("username", None)) and isinstance(username, str):
        password = kwargs.get("password", "")
        headers["Authorization"] = create_basic_authentication(
            username=username, password=password
        )

    return headers if headers else None


class WSClientState(Enum):
    UNINITIALIZED = auto()
    CONNECTING = auto()
    CONNECTED = auto()
    DISCONNECTING = auto()
    DISCONNECTED = auto()


@runtime_checkable
class WSClientListenerC(Protocol):
    async def __call__(self, message: str, **kwargs) -> None:
        ...


@runtime_checkable
class WSClientListenerM(Protocol):
    async def on_message(self, message: str, **kwargs) -> None:
        ...


WSClientListener = Union[WSClientListenerC, WSClientListenerM]


@runtime_checkable
class WSClientProtocol(Protocol):
    # noinspection PyPropertyDefinition
    @property
    def state(self) -> WSClientState:
        ...

    async def connect(self, *, reconnecting: bool = False, **kwargs) -> bool:
        ...

    async def disconnect(self, *, code: int = 1000, reason: str = "", **kwargs) -> None:
        ...

    async def send(self, message: str, **kwargs) -> None:
        ...

    async def on_connect(self, *, reconnecting: bool = False, **kwargs) -> None:
        ...

    async def on_disconnect(
        self, *, code: int = 1000, reason: str = "", **kwargs
    ) -> None:
        ...

    async def on_message(self, message: str, **kwargs) -> None:
        ...


class WSClient:
    RECONNECT_ATTEMPTS_DATA_KEY: str = "reconnect_attempts"

    def __init__(
        self,
        **kwargs,
    ) -> None:
        self._state: WSClientState = WSClientState.UNINITIALIZED
        self._data: Dict[str, Any] = {}
        self._connection = None
        self._send_queue: Optional[asyncio.Queue] = None
        self._recv_queue: Optional[asyncio.Queue] = None
        self._recv_queue: Optional[asyncio.Queue] = None
        self._send_future: Optional[asyncio.Task] = None
        self._recv_future: Optional[asyncio.Task] = None
        self._mesg_future: Optional[asyncio.Task] = None
        self._init_kwargs: Dict[str, Any] = kwargs

        self._conn_event: asyncio.Event = asyncio.Event()
        self._conn_future: asyncio.Task = asyncio.ensure_future(self._conn_task())

        self.listeners: List[WSClientListener] = []

    @property
    def state(self) -> WSClientState:
        return self._state

    async def connect(self, *, reconnecting: bool = False, **kwargs) -> bool:
        if self._state not in (WSClientState.UNINITIALIZED, WSClientState.DISCONNECTED):
            return False

        connect_kwargs: Dict[str, Any] = deepcopy(self._init_kwargs)
        connect_kwargs.update(kwargs)

        url: str = connect_kwargs.pop("url", "")
        if not url:
            url = connect_kwargs.pop("uri", "")  # backwards compatibility
        if not url:
            raise ValueError("Missing or invalid 'url' (or 'uri') keyword argument.")

        url = self._make_url(url, **connect_kwargs)
        headers = self._make_headers(**connect_kwargs)

        self._state = WSClientState.CONNECTING
        self._connection = await self._create_connection(url=url, headers=headers)
        self._setup()
        self._state = WSClientState.CONNECTED

        await self.on_connect(reconnecting=reconnecting, **kwargs)

        return True

    async def disconnect(self, *, code: int = 1000, reason: str = "", **kwargs) -> None:
        if self._state != WSClientState.CONNECTED:
            return

        self._state = WSClientState.DISCONNECTING
        self._tear_down()
        await self._close_connection(code=code, reason=reason)
        self._connection = None
        self._state = WSClientState.DISCONNECTED

        await self.on_disconnect(code=code, reason=reason)

    async def send(self, message: str, **kwargs) -> None:
        await self._send_queue.put(message)

    async def on_connect(self, *, reconnecting: bool = False, **kwargs) -> None:
        self.set_data(self.RECONNECT_ATTEMPTS_DATA_KEY, 0)

    async def on_disconnect(
        self, *, code: int = 1000, reason: str = "", **kwargs
    ) -> None:
        if self.should_reconnect(code=code, reason=reason):
            self._set_reconnect_conn_event_data(
                code=code, reason=reason, number_of_attempts=0
            )
            self._conn_event.set()
        else:
            self.clear_data()

    async def on_message(self, message: str, **kwargs) -> None:
        for listener in self.listeners:
            if isinstance(listener, WSClientListenerC):
                await listener(message, **kwargs)
            elif isinstance(listener, WSClientListenerM):
                await listener.on_message(message, **kwargs)
            else:
                raise TypeError(listener)

    # noinspection PyMethodMayBeStatic
    def should_reconnect(
        self, code: int, reason: str = "", number_of_attempts: int = 0
    ) -> bool:
        return 1001 <= code <= 2999

    # noinspection PyMethodMayBeStatic
    def reconnect_delay(self, number_of_attempts: int) -> float:
        interval: float = 1.0
        backoff_rate: float = 2.0
        return interval * math.pow(backoff_rate, number_of_attempts)

    def get_data(self, key: str) -> Any:
        return self._data.get(key)

    def has_data(self, key: str) -> bool:
        return key in self._data

    def set_data(self, key: str, value: Any) -> None:
        self._data[key] = value

    def set_data_default(self, key: str, default: Any) -> Any:
        return self._data.setdefault(key, default)

    def clear_data(self) -> None:
        self._data.clear()

    def _setup(self) -> None:
        self._send_queue = asyncio.Queue()
        self._recv_queue = asyncio.Queue()
        self._send_future = asyncio.ensure_future(self._send_task())
        self._recv_future = asyncio.ensure_future(self._recv_task())
        self._mesg_future = asyncio.ensure_future(self._mesg_task())

    def _tear_down(self) -> None:
        self._mesg_future.cancel()
        self._recv_future.cancel()
        self._send_future.cancel()
        empty_queue(self._recv_queue)
        empty_queue(self._send_queue)
        self._mesg_future = None
        self._recv_future = None
        self._send_future = None
        self._recv_queue = None
        self._send_queue = None

    # noinspection PyMethodMayBeStatic
    def _make_url(self, url: str, **kwargs) -> str:
        return make_url(url=url, **kwargs)

    # noinspection PyMethodMayBeStatic
    def _make_headers(self, **kwargs) -> Dict[str, str]:
        return make_headers(**kwargs)

    # noinspection PyMethodMayBeStatic
    async def _create_connection(
        self, url: str, headers: Optional[Dict[str, str]], **kwargs
    ) -> Any:
        # pylint: disable=no-member
        return await websockets.connect(uri=url, extra_headers=headers)

    async def _close_connection(
        self, code: int = 1000, reason: str = "", **kwargs
    ) -> None:
        await self._connection.close(code=code, reason=reason)

    async def _conn_task(self) -> None:
        while True:
            await self._conn_event.wait()
            self._conn_event.clear()

            data = getattr(self._conn_event, "data", None)
            if not data:
                continue

            type = data.get("type", "")
            code = data.get("code", 1000)
            reason = data.get("reason", "")
            number_of_attempts = data.get("number_of_attempts", 0)
            self._clear_conn_event_data()

            if type == "disconnect":
                await self.disconnect(code=code, reason=reason)
            elif type == "reconnect":
                delay = self.reconnect_delay(number_of_attempts=number_of_attempts)
                await asyncio.sleep(delay=delay)
                did_reconnect = await self.connect(reconnecting=True)
                if not did_reconnect:
                    if self.should_reconnect(
                        code=code,
                        reason=reason,
                        number_of_attempts=number_of_attempts + 1,
                    ):
                        self._set_reconnect_conn_event_data(
                            code=code,
                            reason=reason,
                            number_of_attempts=number_of_attempts + 1,
                        )
                        self._conn_event.set()
                    else:
                        self.clear_data()
            else:
                raise ValueError(f"unknown event type: {type}")

    async def _send_task(self) -> None:
        while True:
            if self._connection is not None and self._send_queue is not None:
                message = await self._send_queue.get()
                try:
                    await self._connection.send(message)
                except websockets.exceptions.ConnectionClosed as closed:
                    self._set_disconnect_conn_event_data_from_exception(closed=closed)
                    self._conn_event.set()
            else:
                await asyncio.sleep(0.1)

    async def _recv_task(self) -> None:
        while True:
            if self._connection is not None and self._recv_queue is not None:
                try:
                    message = await self._connection.recv()
                    if isinstance(message, bytes):
                        message = message.decode(encoding="utf-8")
                    await self._recv_queue.put(message)
                except websockets.exceptions.ConnectionClosed as closed:
                    self._set_disconnect_conn_event_data_from_exception(closed=closed)
                    self._conn_event.set()
            else:
                await asyncio.sleep(0.1)

    async def _mesg_task(self) -> None:
        while True:
            if self._recv_queue is not None:
                message = await self._recv_queue.get()
                await self.on_message(message=message)
            else:
                await asyncio.sleep(0.1)

    def _set_disconnect_conn_event_data(self, code: int, reason: str = "") -> None:
        setattr(
            self._conn_event,
            "data",
            {
                "type": "disconnect",
                "code": code,
                "reason": reason,
            },
        )

    def _set_disconnect_conn_event_data_from_exception(
        self, closed: websockets.exceptions.ConnectionClosed
    ) -> None:
        if closed.sent is not None:
            source = closed.sent
        elif closed.rcvd is not None:
            source = closed.rcvd
        else:
            source = closed
        self._set_disconnect_conn_event_data(code=source.code, reason=source.reason)

    def _set_reconnect_conn_event_data(
        self, code: int, reason: str = "", number_of_attempts: int = 0
    ) -> None:
        setattr(
            self._conn_event,
            "data",
            {
                "type": "reconnect",
                "code": code,
                "reason": reason,
                "number_of_attempts": number_of_attempts,
            },
        )

    def _clear_conn_event_data(self) -> None:
        setattr(self._conn_event, "data", None)


assert isinstance(WSClient, WSClientProtocol)


__all__ = [
    "WSClient",
    "WSClientListener",
    "WSClientListenerC",
    "WSClientListenerM",
    "WSClientProtocol",
    "WSClientState",
]



# Copyright (c) 2021 AccelByte Inc. All Rights Reserved.
# This is licensed software from AccelByte Inc, for limitations
# and restrictions contact your company contract manager.
#
# Code generated. DO NOT EDIT!

# template file: ags_py_codegen

# pylint: disable=duplicate-code
# pylint: disable=line-too-long
# pylint: disable=missing-function-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=missing-module-docstring
# pylint: disable=too-many-arguments
# pylint: disable=too-many-branches
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-lines
# pylint: disable=too-many-locals
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-statements
# pylint: disable=unused-import

from typing import Any, Dict, List, Optional, Tuple, Union

from ....core import HeaderStr
from ....core import get_namespace as get_services_namespace
from ....core import run_request
from ....core import run_request_async
from ....core import same_doc_as

from ..models import ApiImageDetails
from ..models import ApiImageList
from ..models import ApiImageStorage
from ..models import ApiImageUpdate
from ..models import ResponseErrorResponse

from ..operations.images import ImageGet
from ..operations.images import ImageList
from ..operations.images import ImageMarkForDeletion
from ..operations.images import ImagePatch
from ..operations.images import ImageUnmarkForDeletion
from ..operations.images import ImagesStorage


@same_doc_as(ImageGet)
def image_get(
    image_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """get image details. (ImageGet)

    Required Permission: ADMIN:NAMESPACE:{namespace}:AMS:IMAGE [READ]

    Properties:
        url: /ams/v1/admin/namespaces/{namespace}/images/{imageID}

        method: GET

        tags: ["Images"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        image_id: (imageID) REQUIRED str in path

        namespace: (namespace) REQUIRED str in path

    Responses:
        200: OK - ApiImageDetails (success)

        401: Unauthorized - ResponseErrorResponse (no authorization provided)

        403: Forbidden - ResponseErrorResponse (insufficient permissions)

        404: Not Found - ResponseErrorResponse (account not linked)

        500: Internal Server Error - ResponseErrorResponse (internal server error)
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = ImageGet.create(
        image_id=image_id,
        namespace=namespace,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@same_doc_as(ImageGet)
async def image_get_async(
    image_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """get image details. (ImageGet)

    Required Permission: ADMIN:NAMESPACE:{namespace}:AMS:IMAGE [READ]

    Properties:
        url: /ams/v1/admin/namespaces/{namespace}/images/{imageID}

        method: GET

        tags: ["Images"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        image_id: (imageID) REQUIRED str in path

        namespace: (namespace) REQUIRED str in path

    Responses:
        200: OK - ApiImageDetails (success)

        401: Unauthorized - ResponseErrorResponse (no authorization provided)

        403: Forbidden - ResponseErrorResponse (insufficient permissions)

        404: Not Found - ResponseErrorResponse (account not linked)

        500: Internal Server Error - ResponseErrorResponse (internal server error)
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = ImageGet.create(
        image_id=image_id,
        namespace=namespace,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@same_doc_as(ImageList)
def image_list(
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """get a list of existing images (ImageList)

    Returns images which exist (uploaded, uploading, or building) in the linked account. This route fails if no account is linked

    Required Permission: ADMIN:NAMESPACE:{namespace}:AMS:IMAGE [READ]

    Properties:
        url: /ams/v1/admin/namespaces/{namespace}/images

        method: GET

        tags: ["Images"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        namespace: (namespace) REQUIRED str in path

    Responses:
        200: OK - ApiImageList (success)

        401: Unauthorized - ResponseErrorResponse (no authorization provided)

        403: Forbidden - ResponseErrorResponse (insufficient permissions)

        404: Not Found - ResponseErrorResponse (account not linked)

        500: Internal Server Error - ResponseErrorResponse (internal server error)
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = ImageList.create(
        namespace=namespace,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@same_doc_as(ImageList)
async def image_list_async(
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """get a list of existing images (ImageList)

    Returns images which exist (uploaded, uploading, or building) in the linked account. This route fails if no account is linked

    Required Permission: ADMIN:NAMESPACE:{namespace}:AMS:IMAGE [READ]

    Properties:
        url: /ams/v1/admin/namespaces/{namespace}/images

        method: GET

        tags: ["Images"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        namespace: (namespace) REQUIRED str in path

    Responses:
        200: OK - ApiImageList (success)

        401: Unauthorized - ResponseErrorResponse (no authorization provided)

        403: Forbidden - ResponseErrorResponse (insufficient permissions)

        404: Not Found - ResponseErrorResponse (account not linked)

        500: Internal Server Error - ResponseErrorResponse (internal server error)
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = ImageList.create(
        namespace=namespace,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@same_doc_as(ImageMarkForDeletion)
def image_mark_for_deletion(
    image_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """mark the image for deletion (ImageMarkForDeletion)

    Marks an image for deletion. The image will stop being available for fleets and will eventually be deleted.

    Required Permission: ADMIN:NAMESPACE:{namespace}:AMS:IMAGE [DELETE]

    Properties:
        url: /ams/v1/admin/namespaces/{namespace}/images/{imageID}

        method: DELETE

        tags: ["Images"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        image_id: (imageID) REQUIRED str in path

        namespace: (namespace) REQUIRED str in path

    Responses:
        202: Accepted - (success)

        401: Unauthorized - ResponseErrorResponse (no authorization provided)

        403: Forbidden - ResponseErrorResponse (insufficient permissions)

        404: Not Found - ResponseErrorResponse (image doesn't exist)

        412: Precondition Failed - ResponseErrorResponse (image is being used by 1 or more fleets)

        500: Internal Server Error - ResponseErrorResponse (internal server error)
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = ImageMarkForDeletion.create(
        image_id=image_id,
        namespace=namespace,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@same_doc_as(ImageMarkForDeletion)
async def image_mark_for_deletion_async(
    image_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """mark the image for deletion (ImageMarkForDeletion)

    Marks an image for deletion. The image will stop being available for fleets and will eventually be deleted.

    Required Permission: ADMIN:NAMESPACE:{namespace}:AMS:IMAGE [DELETE]

    Properties:
        url: /ams/v1/admin/namespaces/{namespace}/images/{imageID}

        method: DELETE

        tags: ["Images"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        image_id: (imageID) REQUIRED str in path

        namespace: (namespace) REQUIRED str in path

    Responses:
        202: Accepted - (success)

        401: Unauthorized - ResponseErrorResponse (no authorization provided)

        403: Forbidden - ResponseErrorResponse (insufficient permissions)

        404: Not Found - ResponseErrorResponse (image doesn't exist)

        412: Precondition Failed - ResponseErrorResponse (image is being used by 1 or more fleets)

        500: Internal Server Error - ResponseErrorResponse (internal server error)
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = ImageMarkForDeletion.create(
        image_id=image_id,
        namespace=namespace,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@same_doc_as(ImagePatch)
def image_patch(
    body: ApiImageUpdate,
    image_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """edit the image (ImagePatch)

    This allows editing of the image name, toggling `IsProtected`, or adding & removal of tags

    Required Permission: ADMIN:NAMESPACE:{namespace}:AMS:IMAGE [UPDATE]

    Properties:
        url: /ams/v1/admin/namespaces/{namespace}/images/{imageID}

        method: PATCH

        tags: ["Images"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        body: (body) REQUIRED ApiImageUpdate in body

        image_id: (imageID) REQUIRED str in path

        namespace: (namespace) REQUIRED str in path

    Responses:
        200: OK - ApiImageDetails (success)

        401: Unauthorized - ResponseErrorResponse (no authorization provided)

        403: Forbidden - ResponseErrorResponse (insufficient permissions)

        404: Not Found - ResponseErrorResponse (account not linked)

        500: Internal Server Error - ResponseErrorResponse (internal server error)
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = ImagePatch.create(
        body=body,
        image_id=image_id,
        namespace=namespace,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@same_doc_as(ImagePatch)
async def image_patch_async(
    body: ApiImageUpdate,
    image_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """edit the image (ImagePatch)

    This allows editing of the image name, toggling `IsProtected`, or adding & removal of tags

    Required Permission: ADMIN:NAMESPACE:{namespace}:AMS:IMAGE [UPDATE]

    Properties:
        url: /ams/v1/admin/namespaces/{namespace}/images/{imageID}

        method: PATCH

        tags: ["Images"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        body: (body) REQUIRED ApiImageUpdate in body

        image_id: (imageID) REQUIRED str in path

        namespace: (namespace) REQUIRED str in path

    Responses:
        200: OK - ApiImageDetails (success)

        401: Unauthorized - ResponseErrorResponse (no authorization provided)

        403: Forbidden - ResponseErrorResponse (insufficient permissions)

        404: Not Found - ResponseErrorResponse (account not linked)

        500: Internal Server Error - ResponseErrorResponse (internal server error)
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = ImagePatch.create(
        body=body,
        image_id=image_id,
        namespace=namespace,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@same_doc_as(ImageUnmarkForDeletion)
def image_unmark_for_deletion(
    image_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """unmarks the image for deletion (ImageUnmarkForDeletion)

    Unmarks an image for deletion. The image will be available for fleets.

    Required Permission: ADMIN:NAMESPACE:{namespace}:AMS:IMAGE [DELETE]

    Properties:
        url: /ams/v1/admin/namespaces/{namespace}/images/{imageID}/restore

        method: POST

        tags: ["Images"]

        consumes: []

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        image_id: (imageID) REQUIRED str in path

        namespace: (namespace) REQUIRED str in path

    Responses:
        202: Accepted - (success)

        401: Unauthorized - ResponseErrorResponse (no authorization provided)

        403: Forbidden - ResponseErrorResponse (insufficient permissions)

        404: Not Found - ResponseErrorResponse (image doesn't exist)

        412: Precondition Failed - ResponseErrorResponse (image wasn't marked for deletion)

        500: Internal Server Error - ResponseErrorResponse (internal server error)
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = ImageUnmarkForDeletion.create(
        image_id=image_id,
        namespace=namespace,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@same_doc_as(ImageUnmarkForDeletion)
async def image_unmark_for_deletion_async(
    image_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """unmarks the image for deletion (ImageUnmarkForDeletion)

    Unmarks an image for deletion. The image will be available for fleets.

    Required Permission: ADMIN:NAMESPACE:{namespace}:AMS:IMAGE [DELETE]

    Properties:
        url: /ams/v1/admin/namespaces/{namespace}/images/{imageID}/restore

        method: POST

        tags: ["Images"]

        consumes: []

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        image_id: (imageID) REQUIRED str in path

        namespace: (namespace) REQUIRED str in path

    Responses:
        202: Accepted - (success)

        401: Unauthorized - ResponseErrorResponse (no authorization provided)

        403: Forbidden - ResponseErrorResponse (insufficient permissions)

        404: Not Found - ResponseErrorResponse (image doesn't exist)

        412: Precondition Failed - ResponseErrorResponse (image wasn't marked for deletion)

        500: Internal Server Error - ResponseErrorResponse (internal server error)
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = ImageUnmarkForDeletion.create(
        image_id=image_id,
        namespace=namespace,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@same_doc_as(ImagesStorage)
def images_storage(
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """get current usage for images storage (ImagesStorage)

    Returns information regarding the account's usage for images storage including the free tier quota

    Required Permission: ADMIN:NAMESPACE:{namespace}:AMS:IMAGE [READ]

    Properties:
        url: /ams/v1/admin/namespaces/{namespace}/images-storage

        method: GET

        tags: ["Images"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        namespace: (namespace) REQUIRED str in path

    Responses:
        200: OK - ApiImageStorage (success)

        401: Unauthorized - ResponseErrorResponse (no authorization provided)

        403: Forbidden - ResponseErrorResponse (insufficient permissions)

        404: Not Found - ResponseErrorResponse (account not linked)

        500: Internal Server Error - ResponseErrorResponse (internal server error)
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = ImagesStorage.create(
        namespace=namespace,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@same_doc_as(ImagesStorage)
async def images_storage_async(
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """get current usage for images storage (ImagesStorage)

    Returns information regarding the account's usage for images storage including the free tier quota

    Required Permission: ADMIN:NAMESPACE:{namespace}:AMS:IMAGE [READ]

    Properties:
        url: /ams/v1/admin/namespaces/{namespace}/images-storage

        method: GET

        tags: ["Images"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        namespace: (namespace) REQUIRED str in path

    Responses:
        200: OK - ApiImageStorage (success)

        401: Unauthorized - ResponseErrorResponse (no authorization provided)

        403: Forbidden - ResponseErrorResponse (insufficient permissions)

        404: Not Found - ResponseErrorResponse (account not linked)

        500: Internal Server Error - ResponseErrorResponse (internal server error)
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = ImagesStorage.create(
        namespace=namespace,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )



# Copyright (c) 2021 AccelByte Inc. All Rights Reserved.
# This is licensed software from AccelByte Inc, for limitations
# and restrictions contact your company contract manager.
#
# Code generated. DO NOT EDIT!

# template file: ags_py_codegen

# AccelByte Gaming Services Iam Service

# pylint: disable=duplicate-code
# pylint: disable=line-too-long
# pylint: disable=missing-function-docstring
# pylint: disable=missing-module-docstring
# pylint: disable=too-many-arguments
# pylint: disable=too-many-branches
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-lines
# pylint: disable=too-many-locals
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-statements
# pylint: disable=unused-import

from __future__ import annotations
from typing import Any, Dict, List, Optional, Tuple, Union

from ....core import Model


class ModelUserPlatformLinkHistory(Model):
    """Model user platform link history (model.UserPlatformLinkHistory)

    Properties:
        action: (action) REQUIRED str

        actor: (actor) REQUIRED str

        actor_type: (actorType) REQUIRED str

        created_at: (createdAt) REQUIRED int

        namespace: (namespace) REQUIRED str

        platform: (platform) REQUIRED str

        platform_display_name: (platformDisplayName) REQUIRED str

        platform_id: (platformId) REQUIRED str

        platform_user_id: (platformUserId) REQUIRED str

        publisher_user_id: (publisherUserId) REQUIRED str
    """

    # region fields

    action: str  # REQUIRED
    actor: str  # REQUIRED
    actor_type: str  # REQUIRED
    created_at: int  # REQUIRED
    namespace: str  # REQUIRED
    platform: str  # REQUIRED
    platform_display_name: str  # REQUIRED
    platform_id: str  # REQUIRED
    platform_user_id: str  # REQUIRED
    publisher_user_id: str  # REQUIRED

    # endregion fields

    # region with_x methods

    def with_action(self, value: str) -> ModelUserPlatformLinkHistory:
        self.action = value
        return self

    def with_actor(self, value: str) -> ModelUserPlatformLinkHistory:
        self.actor = value
        return self

    def with_actor_type(self, value: str) -> ModelUserPlatformLinkHistory:
        self.actor_type = value
        return self

    def with_created_at(self, value: int) -> ModelUserPlatformLinkHistory:
        self.created_at = value
        return self

    def with_namespace(self, value: str) -> ModelUserPlatformLinkHistory:
        self.namespace = value
        return self

    def with_platform(self, value: str) -> ModelUserPlatformLinkHistory:
        self.platform = value
        return self

    def with_platform_display_name(self, value: str) -> ModelUserPlatformLinkHistory:
        self.platform_display_name = value
        return self

    def with_platform_id(self, value: str) -> ModelUserPlatformLinkHistory:
        self.platform_id = value
        return self

    def with_platform_user_id(self, value: str) -> ModelUserPlatformLinkHistory:
        self.platform_user_id = value
        return self

    def with_publisher_user_id(self, value: str) -> ModelUserPlatformLinkHistory:
        self.publisher_user_id = value
        return self

    # endregion with_x methods

    # region to methods

    def to_dict(self, include_empty: bool = False) -> dict:
        result: dict = {}
        if hasattr(self, "action"):
            result["action"] = str(self.action)
        elif include_empty:
            result["action"] = ""
        if hasattr(self, "actor"):
            result["actor"] = str(self.actor)
        elif include_empty:
            result["actor"] = ""
        if hasattr(self, "actor_type"):
            result["actorType"] = str(self.actor_type)
        elif include_empty:
            result["actorType"] = ""
        if hasattr(self, "created_at"):
            result["createdAt"] = int(self.created_at)
        elif include_empty:
            result["createdAt"] = 0
        if hasattr(self, "namespace"):
            result["namespace"] = str(self.namespace)
        elif include_empty:
            result["namespace"] = ""
        if hasattr(self, "platform"):
            result["platform"] = str(self.platform)
        elif include_empty:
            result["platform"] = ""
        if hasattr(self, "platform_display_name"):
            result["platformDisplayName"] = str(self.platform_display_name)
        elif include_empty:
            result["platformDisplayName"] = ""
        if hasattr(self, "platform_id"):
            result["platformId"] = str(self.platform_id)
        elif include_empty:
            result["platformId"] = ""
        if hasattr(self, "platform_user_id"):
            result["platformUserId"] = str(self.platform_user_id)
        elif include_empty:
            result["platformUserId"] = ""
        if hasattr(self, "publisher_user_id"):
            result["publisherUserId"] = str(self.publisher_user_id)
        elif include_empty:
            result["publisherUserId"] = ""
        return result

    # endregion to methods

    # region static methods

    @classmethod
    def create(
        cls,
        action: str,
        actor: str,
        actor_type: str,
        created_at: int,
        namespace: str,
        platform: str,
        platform_display_name: str,
        platform_id: str,
        platform_user_id: str,
        publisher_user_id: str,
        **kwargs,
    ) -> ModelUserPlatformLinkHistory:
        instance = cls()
        instance.action = action
        instance.actor = actor
        instance.actor_type = actor_type
        instance.created_at = created_at
        instance.namespace = namespace
        instance.platform = platform
        instance.platform_display_name = platform_display_name
        instance.platform_id = platform_id
        instance.platform_user_id = platform_user_id
        instance.publisher_user_id = publisher_user_id
        return instance

    @classmethod
    def create_from_dict(
        cls, dict_: dict, include_empty: bool = False
    ) -> ModelUserPlatformLinkHistory:
        instance = cls()
        if not dict_:
            return instance
        if "action" in dict_ and dict_["action"] is not None:
            instance.action = str(dict_["action"])
        elif include_empty:
            instance.action = ""
        if "actor" in dict_ and dict_["actor"] is not None:
            instance.actor = str(dict_["actor"])
        elif include_empty:
            instance.actor = ""
        if "actorType" in dict_ and dict_["actorType"] is not None:
            instance.actor_type = str(dict_["actorType"])
        elif include_empty:
            instance.actor_type = ""
        if "createdAt" in dict_ and dict_["createdAt"] is not None:
            instance.created_at = int(dict_["createdAt"])
        elif include_empty:
            instance.created_at = 0
        if "namespace" in dict_ and dict_["namespace"] is not None:
            instance.namespace = str(dict_["namespace"])
        elif include_empty:
            instance.namespace = ""
        if "platform" in dict_ and dict_["platform"] is not None:
            instance.platform = str(dict_["platform"])
        elif include_empty:
            instance.platform = ""
        if "platformDisplayName" in dict_ and dict_["platformDisplayName"] is not None:
            instance.platform_display_name = str(dict_["platformDisplayName"])
        elif include_empty:
            instance.platform_display_name = ""
        if "platformId" in dict_ and dict_["platformId"] is not None:
            instance.platform_id = str(dict_["platformId"])
        elif include_empty:
            instance.platform_id = ""
        if "platformUserId" in dict_ and dict_["platformUserId"] is not None:
            instance.platform_user_id = str(dict_["platformUserId"])
        elif include_empty:
            instance.platform_user_id = ""
        if "publisherUserId" in dict_ and dict_["publisherUserId"] is not None:
            instance.publisher_user_id = str(dict_["publisherUserId"])
        elif include_empty:
            instance.publisher_user_id = ""
        return instance

    @classmethod
    def create_many_from_dict(
        cls, dict_: dict, include_empty: bool = False
    ) -> Dict[str, ModelUserPlatformLinkHistory]:
        return (
            {k: cls.create_from_dict(v, include_empty=include_empty) for k, v in dict_}
            if dict_
            else {}
        )

    @classmethod
    def create_many_from_list(
        cls, list_: list, include_empty: bool = False
    ) -> List[ModelUserPlatformLinkHistory]:
        return (
            [cls.create_from_dict(i, include_empty=include_empty) for i in list_]
            if list_
            else []
        )

    @classmethod
    def create_from_any(
        cls, any_: any, include_empty: bool = False, many: bool = False
    ) -> Union[
        ModelUserPlatformLinkHistory,
        List[ModelUserPlatformLinkHistory],
        Dict[Any, ModelUserPlatformLinkHistory],
    ]:
        if many:
            if isinstance(any_, dict):
                return cls.create_many_from_dict(any_, include_empty=include_empty)
            elif isinstance(any_, list):
                return cls.create_many_from_list(any_, include_empty=include_empty)
            else:
                raise ValueError()
        else:
            return cls.create_from_dict(any_, include_empty=include_empty)

    @staticmethod
    def get_field_info() -> Dict[str, str]:
        return {
            "action": "action",
            "actor": "actor",
            "actorType": "actor_type",
            "createdAt": "created_at",
            "namespace": "namespace",
            "platform": "platform",
            "platformDisplayName": "platform_display_name",
            "platformId": "platform_id",
            "platformUserId": "platform_user_id",
            "publisherUserId": "publisher_user_id",
        }

    @staticmethod
    def get_required_map() -> Dict[str, bool]:
        return {
            "action": True,
            "actor": True,
            "actorType": True,
            "createdAt": True,
            "namespace": True,
            "platform": True,
            "platformDisplayName": True,
            "platformId": True,
            "platformUserId": True,
            "publisherUserId": True,
        }

    # endregion static methods



# Copyright (c) 2021 AccelByte Inc. All Rights Reserved.
# This is licensed software from AccelByte Inc, for limitations
# and restrictions contact your company contract manager.
#
# Code generated. DO NOT EDIT!

# template file: ags_py_codegen

"""Auto-generated package that contains models used by the AccelByte Gaming Services Lobby Server."""

__version__ = "3.36.0"
__author__ = "AccelByte"
__email__ = "dev@accelbyte.net"

# pylint: disable=line-too-long

from .admin_bulk_block_players_v1 import AdminBulkBlockPlayersV1
from .admin_bulk_unblock_players_v1 import AdminBulkUnblockPlayersV1
from .admin_get_all_player_se_b925d7 import AdminGetAllPlayerSessionAttribute
from .admin_get_bulk_player_b_0addc5 import AdminGetBulkPlayerBlockedPlayersV1
from .admin_get_lobby_ccu import AdminGetLobbyCCU
from .admin_get_player_blocke_8955db import AdminGetPlayerBlockedByPlayersV1
from .admin_get_player_blocke_7da3f3 import AdminGetPlayerBlockedPlayersV1
from .admin_get_player_sessio_625ce8 import AdminGetPlayerSessionAttribute
from .admin_set_player_sessio_1bc722 import AdminSetPlayerSessionAttribute
from .public_get_player_block_0dc6b7 import PublicGetPlayerBlockedByPlayersV1
from .public_get_player_block_55d58a import PublicGetPlayerBlockedPlayersV1
from .public_player_block_players_v1 import PublicPlayerBlockPlayersV1
from .public_unblock_player_v1 import PublicUnblockPlayerV1



# Copyright (c) 2021 AccelByte Inc. All Rights Reserved.
# This is licensed software from AccelByte Inc, for limitations
# and restrictions contact your company contract manager.
#
# Code generated. DO NOT EDIT!

# template file: ags_py_codegen

# pylint: disable=duplicate-code
# pylint: disable=line-too-long
# pylint: disable=missing-function-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=missing-module-docstring
# pylint: disable=too-many-arguments
# pylint: disable=too-many-branches
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-lines
# pylint: disable=too-many-locals
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-statements
# pylint: disable=unused-import

from typing import Any, Dict, List, Optional, Tuple, Union

from ....core import HeaderStr
from ....core import get_namespace as get_services_namespace
from ....core import run_request
from ....core import run_request_async
from ....core import same_doc_as

from ..models import RestapiAdminAllReasonsResponse
from ..models import RestapiAdminReasonListResponse
from ..models import RestapiAdminReasonResponse
from ..models import RestapiCreateReasonGroupRequest
from ..models import RestapiCreateReasonRequest
from ..models import RestapiErrorResponse
from ..models import RestapiReasonGroupListResponse
from ..models import RestapiReasonGroupResponse
from ..models import RestapiUnusedReasonListResponse
from ..models import RestapiUpdateReasonGroupRequest

from ..operations.admin_reasons import AdminGetAllReasons
from ..operations.admin_reasons import AdminGetReason
from ..operations.admin_reasons import AdminGetReasons
from ..operations.admin_reasons import AdminGetUnusedReasons
from ..operations.admin_reasons import AdminListReasonGroups
from ..operations.admin_reasons import CreateReason
from ..operations.admin_reasons import CreateReasonGroup
from ..operations.admin_reasons import DeleteReason
from ..operations.admin_reasons import DeleteReasonGroup
from ..operations.admin_reasons import GetReasonGroup
from ..operations.admin_reasons import UpdateReason
from ..operations.admin_reasons import UpdateReasonGroup


@same_doc_as(AdminGetAllReasons)
def admin_get_all_reasons(
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Get all reasons (adminGetAllReasons)

    This endpoint get all reasons without pagination.

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasons/all

        method: GET

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        namespace: (namespace) REQUIRED str in path

    Responses:
        200: OK - RestapiAdminAllReasonsResponse (OK)

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = AdminGetAllReasons.create(
        namespace=namespace,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@same_doc_as(AdminGetAllReasons)
async def admin_get_all_reasons_async(
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Get all reasons (adminGetAllReasons)

    This endpoint get all reasons without pagination.

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasons/all

        method: GET

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        namespace: (namespace) REQUIRED str in path

    Responses:
        200: OK - RestapiAdminAllReasonsResponse (OK)

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = AdminGetAllReasons.create(
        namespace=namespace,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@same_doc_as(AdminGetReason)
def admin_get_reason(
    reason_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Get a single reason (adminGetReason)

    This endpoint get a single reason.

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasons/{reasonId}

        method: GET

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        namespace: (namespace) REQUIRED str in path

        reason_id: (reasonId) REQUIRED str in path

    Responses:
        200: OK - RestapiAdminReasonResponse (OK)

        404: Not Found - RestapiErrorResponse

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = AdminGetReason.create(
        reason_id=reason_id,
        namespace=namespace,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@same_doc_as(AdminGetReason)
async def admin_get_reason_async(
    reason_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Get a single reason (adminGetReason)

    This endpoint get a single reason.

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasons/{reasonId}

        method: GET

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        namespace: (namespace) REQUIRED str in path

        reason_id: (reasonId) REQUIRED str in path

    Responses:
        200: OK - RestapiAdminReasonResponse (OK)

        404: Not Found - RestapiErrorResponse

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = AdminGetReason.create(
        reason_id=reason_id,
        namespace=namespace,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@same_doc_as(AdminGetReasons)
def admin_get_reasons(
    group: Optional[str] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    title: Optional[str] = None,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Get list of reasons (adminGetReasons)

    This endpoint get reasons with pagination

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasons

        method: GET

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        namespace: (namespace) REQUIRED str in path

        group: (group) OPTIONAL str in query

        limit: (limit) OPTIONAL int in query

        offset: (offset) OPTIONAL int in query

        title: (title) OPTIONAL str in query

    Responses:
        200: OK - RestapiAdminReasonListResponse (OK)

        404: Not Found - RestapiErrorResponse

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = AdminGetReasons.create(
        group=group,
        limit=limit,
        offset=offset,
        title=title,
        namespace=namespace,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@same_doc_as(AdminGetReasons)
async def admin_get_reasons_async(
    group: Optional[str] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    title: Optional[str] = None,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Get list of reasons (adminGetReasons)

    This endpoint get reasons with pagination

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasons

        method: GET

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        namespace: (namespace) REQUIRED str in path

        group: (group) OPTIONAL str in query

        limit: (limit) OPTIONAL int in query

        offset: (offset) OPTIONAL int in query

        title: (title) OPTIONAL str in query

    Responses:
        200: OK - RestapiAdminReasonListResponse (OK)

        404: Not Found - RestapiErrorResponse

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = AdminGetReasons.create(
        group=group,
        limit=limit,
        offset=offset,
        title=title,
        namespace=namespace,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@same_doc_as(AdminGetUnusedReasons)
def admin_get_unused_reasons(
    category: str,
    extension_category: Optional[str] = None,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Get list of reasons that not used by moderation rules (adminGetUnusedReasons)

    This endpoint get reasons not used by moderation rules.

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasons/unused

        method: GET

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        namespace: (namespace) REQUIRED str in path

        extension_category: (extensionCategory) OPTIONAL str in query

        category: (category) REQUIRED str in query

    Responses:
        200: OK - RestapiUnusedReasonListResponse (OK)

        404: Not Found - RestapiErrorResponse (Not Found)

        500: Internal Server Error - RestapiErrorResponse (Internal Server Error)
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = AdminGetUnusedReasons.create(
        category=category,
        extension_category=extension_category,
        namespace=namespace,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@same_doc_as(AdminGetUnusedReasons)
async def admin_get_unused_reasons_async(
    category: str,
    extension_category: Optional[str] = None,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Get list of reasons that not used by moderation rules (adminGetUnusedReasons)

    This endpoint get reasons not used by moderation rules.

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasons/unused

        method: GET

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        namespace: (namespace) REQUIRED str in path

        extension_category: (extensionCategory) OPTIONAL str in query

        category: (category) REQUIRED str in query

    Responses:
        200: OK - RestapiUnusedReasonListResponse (OK)

        404: Not Found - RestapiErrorResponse (Not Found)

        500: Internal Server Error - RestapiErrorResponse (Internal Server Error)
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = AdminGetUnusedReasons.create(
        category=category,
        extension_category=extension_category,
        namespace=namespace,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@same_doc_as(AdminListReasonGroups)
def admin_list_reason_groups(
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """List reason groups under a namespace (adminListReasonGroups)

    Return list of reason groups ID and title under given namespace.
    To fetch the reasons inside a group, use get reason group endpoint.

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasonGroups

        method: GET

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        namespace: (namespace) REQUIRED str in path

        limit: (limit) OPTIONAL int in query

        offset: (offset) OPTIONAL int in query

    Responses:
        200: OK - RestapiReasonGroupListResponse

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = AdminListReasonGroups.create(
        limit=limit,
        offset=offset,
        namespace=namespace,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@same_doc_as(AdminListReasonGroups)
async def admin_list_reason_groups_async(
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """List reason groups under a namespace (adminListReasonGroups)

    Return list of reason groups ID and title under given namespace.
    To fetch the reasons inside a group, use get reason group endpoint.

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasonGroups

        method: GET

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        namespace: (namespace) REQUIRED str in path

        limit: (limit) OPTIONAL int in query

        offset: (offset) OPTIONAL int in query

    Responses:
        200: OK - RestapiReasonGroupListResponse

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = AdminListReasonGroups.create(
        limit=limit,
        offset=offset,
        namespace=namespace,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@same_doc_as(CreateReason)
def create_reason(
    body: RestapiCreateReasonRequest,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Create a report reason (createReason)

    This endpoint create a reason for a namespace.

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasons

        method: POST

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        body: (body) REQUIRED RestapiCreateReasonRequest in body

        namespace: (namespace) REQUIRED str in path

    Responses:
        201: Created - RestapiAdminReasonResponse

        400: Bad Request - RestapiErrorResponse

        409: Conflict - RestapiErrorResponse (Conflict)

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = CreateReason.create(
        body=body,
        namespace=namespace,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@same_doc_as(CreateReason)
async def create_reason_async(
    body: RestapiCreateReasonRequest,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Create a report reason (createReason)

    This endpoint create a reason for a namespace.

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasons

        method: POST

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        body: (body) REQUIRED RestapiCreateReasonRequest in body

        namespace: (namespace) REQUIRED str in path

    Responses:
        201: Created - RestapiAdminReasonResponse

        400: Bad Request - RestapiErrorResponse

        409: Conflict - RestapiErrorResponse (Conflict)

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = CreateReason.create(
        body=body,
        namespace=namespace,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@same_doc_as(CreateReasonGroup)
def create_reason_group(
    body: RestapiCreateReasonGroupRequest,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Create a reason group (createReasonGroup)

    Create a reason group for easier query. You can query reasons by specifying
    the group title in the list reasons query. Reason group title is case insensitive,
    meaning you can't have **reason** if you already create a reason titled **Reason**

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasonGroups

        method: POST

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        body: (body) REQUIRED RestapiCreateReasonGroupRequest in body

        namespace: (namespace) REQUIRED str in path

    Responses:
        201: Created - RestapiReasonGroupResponse

        400: Bad Request - RestapiErrorResponse

        409: Conflict - RestapiErrorResponse

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = CreateReasonGroup.create(
        body=body,
        namespace=namespace,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@same_doc_as(CreateReasonGroup)
async def create_reason_group_async(
    body: RestapiCreateReasonGroupRequest,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Create a reason group (createReasonGroup)

    Create a reason group for easier query. You can query reasons by specifying
    the group title in the list reasons query. Reason group title is case insensitive,
    meaning you can't have **reason** if you already create a reason titled **Reason**

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasonGroups

        method: POST

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        body: (body) REQUIRED RestapiCreateReasonGroupRequest in body

        namespace: (namespace) REQUIRED str in path

    Responses:
        201: Created - RestapiReasonGroupResponse

        400: Bad Request - RestapiErrorResponse

        409: Conflict - RestapiErrorResponse

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = CreateReasonGroup.create(
        body=body,
        namespace=namespace,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@same_doc_as(DeleteReason)
def delete_reason(
    reason_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Delete a report reason (deleteReason)

    This endpoint delete a reason for a namespace with ID.

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasons/{reasonId}

        method: DELETE

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        namespace: (namespace) REQUIRED str in path

        reason_id: (reasonId) REQUIRED str in path

    Responses:
        204: No Content -

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = DeleteReason.create(
        reason_id=reason_id,
        namespace=namespace,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@same_doc_as(DeleteReason)
async def delete_reason_async(
    reason_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Delete a report reason (deleteReason)

    This endpoint delete a reason for a namespace with ID.

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasons/{reasonId}

        method: DELETE

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        namespace: (namespace) REQUIRED str in path

        reason_id: (reasonId) REQUIRED str in path

    Responses:
        204: No Content -

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = DeleteReason.create(
        reason_id=reason_id,
        namespace=namespace,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@same_doc_as(DeleteReasonGroup)
def delete_reason_group(
    group_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Delete a reason group (deleteReasonGroup)

    This endpoint delete a reason group for a namespace with ID.

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasonGroups/{groupId}

        method: DELETE

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        group_id: (groupId) REQUIRED str in path

        namespace: (namespace) REQUIRED str in path

    Responses:
        204: No Content -

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = DeleteReasonGroup.create(
        group_id=group_id,
        namespace=namespace,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@same_doc_as(DeleteReasonGroup)
async def delete_reason_group_async(
    group_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Delete a reason group (deleteReasonGroup)

    This endpoint delete a reason group for a namespace with ID.

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasonGroups/{groupId}

        method: DELETE

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        group_id: (groupId) REQUIRED str in path

        namespace: (namespace) REQUIRED str in path

    Responses:
        204: No Content -

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = DeleteReasonGroup.create(
        group_id=group_id,
        namespace=namespace,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@same_doc_as(GetReasonGroup)
def get_reason_group(
    group_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Get reason group (getReasonGroup)

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasonGroups/{groupId}

        method: GET

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        group_id: (groupId) REQUIRED str in path

        namespace: (namespace) REQUIRED str in path

    Responses:
        200: OK - RestapiReasonGroupResponse

        404: Not Found - RestapiErrorResponse

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = GetReasonGroup.create(
        group_id=group_id,
        namespace=namespace,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@same_doc_as(GetReasonGroup)
async def get_reason_group_async(
    group_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Get reason group (getReasonGroup)

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasonGroups/{groupId}

        method: GET

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        group_id: (groupId) REQUIRED str in path

        namespace: (namespace) REQUIRED str in path

    Responses:
        200: OK - RestapiReasonGroupResponse

        404: Not Found - RestapiErrorResponse

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = GetReasonGroup.create(
        group_id=group_id,
        namespace=namespace,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@same_doc_as(UpdateReason)
def update_reason(
    body: RestapiCreateReasonRequest,
    reason_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Update a report reason (updateReason)

    This endpoint update a reason for a namespace with ID.

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasons/{reasonId}

        method: PATCH

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        body: (body) REQUIRED RestapiCreateReasonRequest in body

        namespace: (namespace) REQUIRED str in path

        reason_id: (reasonId) REQUIRED str in path

    Responses:
        200: OK - RestapiAdminReasonResponse

        400: Bad Request - RestapiErrorResponse

        409: Conflict - RestapiErrorResponse (Conflict)

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = UpdateReason.create(
        body=body,
        reason_id=reason_id,
        namespace=namespace,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@same_doc_as(UpdateReason)
async def update_reason_async(
    body: RestapiCreateReasonRequest,
    reason_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Update a report reason (updateReason)

    This endpoint update a reason for a namespace with ID.

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasons/{reasonId}

        method: PATCH

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        body: (body) REQUIRED RestapiCreateReasonRequest in body

        namespace: (namespace) REQUIRED str in path

        reason_id: (reasonId) REQUIRED str in path

    Responses:
        200: OK - RestapiAdminReasonResponse

        400: Bad Request - RestapiErrorResponse

        409: Conflict - RestapiErrorResponse (Conflict)

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = UpdateReason.create(
        body=body,
        reason_id=reason_id,
        namespace=namespace,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )


@same_doc_as(UpdateReasonGroup)
def update_reason_group(
    body: RestapiUpdateReasonGroupRequest,
    group_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Update a reason group (updateReasonGroup)

    Reason group title is case insensitive, meaning you can't have **reason** if you already create a reason titled **Reason**
    If no reasonIds passed when updating, the current reasons under the reason group will be kept (reasons will not be removed from the group).

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasonGroups/{groupId}

        method: PATCH

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        body: (body) REQUIRED RestapiUpdateReasonGroupRequest in body

        group_id: (groupId) REQUIRED str in path

        namespace: (namespace) REQUIRED str in path

    Responses:
        200: OK - RestapiReasonGroupResponse

        400: Bad Request - RestapiErrorResponse

        409: Conflict - RestapiErrorResponse

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = UpdateReasonGroup.create(
        body=body,
        group_id=group_id,
        namespace=namespace,
    )
    return run_request(request, additional_headers=x_additional_headers, **kwargs)


@same_doc_as(UpdateReasonGroup)
async def update_reason_group_async(
    body: RestapiUpdateReasonGroupRequest,
    group_id: str,
    namespace: Optional[str] = None,
    x_additional_headers: Optional[Dict[str, str]] = None,
    **kwargs
):
    """Update a reason group (updateReasonGroup)

    Reason group title is case insensitive, meaning you can't have **reason** if you already create a reason titled **Reason**
    If no reasonIds passed when updating, the current reasons under the reason group will be kept (reasons will not be removed from the group).

    Properties:
        url: /reporting/v1/admin/namespaces/{namespace}/reasonGroups/{groupId}

        method: PATCH

        tags: ["Admin Reasons"]

        consumes: ["application/json"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        body: (body) REQUIRED RestapiUpdateReasonGroupRequest in body

        group_id: (groupId) REQUIRED str in path

        namespace: (namespace) REQUIRED str in path

    Responses:
        200: OK - RestapiReasonGroupResponse

        400: Bad Request - RestapiErrorResponse

        409: Conflict - RestapiErrorResponse

        500: Internal Server Error - RestapiErrorResponse
    """
    if namespace is None:
        namespace, error = get_services_namespace()
        if error:
            return None, error
    request = UpdateReasonGroup.create(
        body=body,
        group_id=group_id,
        namespace=namespace,
    )
    return await run_request_async(
        request, additional_headers=x_additional_headers, **kwargs
    )



# Copyright (c) 2021 AccelByte Inc. All Rights Reserved.
# This is licensed software from AccelByte Inc, for limitations
# and restrictions contact your company contract manager.
#
# Code generated. DO NOT EDIT!

# template file: ags_py_codegen

# pylint: disable=duplicate-code
# pylint: disable=line-too-long
# pylint: disable=missing-function-docstring
# pylint: disable=missing-module-docstring
# pylint: disable=too-many-arguments
# pylint: disable=too-many-branches
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-lines
# pylint: disable=too-many-locals
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-statements
# pylint: disable=unused-import

# AccelByte Gaming Services Iam Service

from __future__ import annotations
from typing import Any, Dict, List, Optional, Tuple, Union

from .....core import Operation
from .....core import HeaderStr
from .....core import HttpResponse
from .....core import deprecated

from ...models import RestErrorResponse


class PlatformLink(Operation):
    """Link user's account with platform (PlatformLink)

    ## The endpoint is going to be deprecated
    ### Endpoint migration guide
    - **Substitute endpoint: _/iam/v3/public/namespaces/{namespace}/users/me/platforms/{platformId} [POST]_**

    **Prerequisite:** Platform client configuration need to be added to database for specific platformId. Namespace service URL need to be specified (refer to required environment variables).

    ## Supported platforms:
    - **steam**: The ticketâs value is the authentication code returned by Steam.
    - **steamopenid**: Steam's user authentication method using OpenID 2.0. The ticket's value is URL generated by Steam on web authentication
    - **facebook**: The ticketâs value is the authorization code returned by Facebook OAuth
    - **google**: The ticketâs value is the authorization code returned by Google OAuth
    - **oculus**: The ticketâs value is a string composed of Oculus's user ID and the nonce separated by a colon (:).
    - **twitch**: The ticketâs value is the authorization code returned by Twitch OAuth.
    - **android**: The ticket's value is the Androidâs device ID
    - **ios**: The ticket's value is the iOSâs device ID.
    - **apple**: The ticketâs value is the authorization code returned by Apple OAuth.
    - **device**: Every device that doesânt run Android and iOS is categorized as a device platform. The ticket's value is the deviceâs ID.
    - **discord**: The ticketâs value is the authorization code returned by Discord OAuth.

    Properties:
        url: /iam/namespaces/{namespace}/users/{userId}/platforms/{platformId}/link

        method: POST

        tags: ["Users"]

        consumes: ["application/x-www-form-urlencoded"]

        produces: ["application/json"]

        securities: [BEARER_AUTH]

        ticket: (ticket) REQUIRED str in form_data

        namespace: (namespace) REQUIRED str in path

        platform_id: (platformId) REQUIRED str in path

        user_id: (userId) REQUIRED str in path

    Responses:
        204: No Content - (Operation succeeded)

        400: Bad Request - (20019: unable to parse request body)

        401: Unauthorized - RestErrorResponse (20001: unauthorized access)

        403: Forbidden - RestErrorResponse (20013: insufficient permissions)

        404: Not Found - (20008: user not found)

        409: Conflict - (10163: platform is already linked with the user account)

        500: Internal Server Error - (20000: internal server error)
    """

    # region fields

    _url: str = "/iam/namespaces/{namespace}/users/{userId}/platforms/{platformId}/link"
    _method: str = "POST"
    _consumes: List[str] = ["application/x-www-form-urlencoded"]
    _produces: List[str] = ["application/json"]
    _securities: List[List[str]] = [["BEARER_AUTH"]]
    _location_query: str = None

    ticket: str  # REQUIRED in [form_data]
    namespace: str  # REQUIRED in [path]
    platform_id: str  # REQUIRED in [path]
    user_id: str  # REQUIRED in [path]

    # endregion fields

    # region properties

    @property
    def url(self) -> str:
        return self._url

    @property
    def method(self) -> str:
        return self._method

    @property
    def consumes(self) -> List[str]:
        return self._consumes

    @property
    def produces(self) -> List[str]:
        return self._produces

    @property
    def securities(self) -> List[List[str]]:
        return self._securities

    @property
    def location_query(self) -> str:
        return self._location_query

    # endregion properties

    # region get methods

    # endregion get methods

    # region get_x_params methods

    def get_all_params(self) -> dict:
        return {
            "form_data": self.get_form_data_params(),
            "path": self.get_path_params(),
        }

    def get_form_data_params(self) -> dict:
        result = {}
        if hasattr(self, "ticket"):
            result["ticket"] = self.ticket
        return result

    def get_path_params(self) -> dict:
        result = {}
        if hasattr(self, "namespace"):
            result["namespace"] = self.namespace
        if hasattr(self, "platform_id"):
            result["platformId"] = self.platform_id
        if hasattr(self, "user_id"):
            result["userId"] = self.user_id
        return result

    # endregion get_x_params methods

    # region is/has methods

    # endregion is/has methods

    # region with_x methods

    def with_ticket(self, value: str) -> PlatformLink:
        self.ticket = value
        return self

    def with_namespace(self, value: str) -> PlatformLink:
        self.namespace = value
        return self

    def with_platform_id(self, value: str) -> PlatformLink:
        self.platform_id = value
        return self

    def with_user_id(self, value: str) -> PlatformLink:
        self.user_id = value
        return self

    # endregion with_x methods

    # region to methods

    def to_dict(self, include_empty: bool = False) -> dict:
        result: dict = {}
        if hasattr(self, "ticket") and self.ticket:
            result["ticket"] = str(self.ticket)
        elif include_empty:
            result["ticket"] = ""
        if hasattr(self, "namespace") and self.namespace:
            result["namespace"] = str(self.namespace)
        elif include_empty:
            result["namespace"] = ""
        if hasattr(self, "platform_id") and self.platform_id:
            result["platformId"] = str(self.platform_id)
        elif include_empty:
            result["platformId"] = ""
        if hasattr(self, "user_id") and self.user_id:
            result["userId"] = str(self.user_id)
        elif include_empty:
            result["userId"] = ""
        return result

    # endregion to methods

    # region response methods

    # noinspection PyMethodMayBeStatic
    def parse_response(
        self, code: int, content_type: str, content: Any
    ) -> Tuple[None, Union[None, HttpResponse, RestErrorResponse]]:
        """Parse the given response.

        204: No Content - (Operation succeeded)

        400: Bad Request - (20019: unable to parse request body)

        401: Unauthorized - RestErrorResponse (20001: unauthorized access)

        403: Forbidden - RestErrorResponse (20013: insufficient permissions)

        404: Not Found - (20008: user not found)

        409: Conflict - (10163: platform is already linked with the user account)

        500: Internal Server Error - (20000: internal server error)

        ---: HttpResponse (Undocumented Response)

        ---: HttpResponse (Unexpected Content-Type Error)

        ---: HttpResponse (Unhandled Error)
        """
        pre_processed_response, error = self.pre_process_response(
            code=code, content_type=content_type, content=content
        )
        if error is not None:
            return None, None if error.is_no_content() else error
        code, content_type, content = pre_processed_response

        if code == 204:
            return None, None
        if code == 400:
            return None, HttpResponse.create(code, "Bad Request")
        if code == 401:
            return None, RestErrorResponse.create_from_dict(content)
        if code == 403:
            return None, RestErrorResponse.create_from_dict(content)
        if code == 404:
            return None, HttpResponse.create(code, "Not Found")
        if code == 409:
            return None, HttpResponse.create(code, "Conflict")
        if code == 500:
            return None, HttpResponse.create(code, "Internal Server Error")

        return self.handle_undocumented_response(
            code=code, content_type=content_type, content=content
        )

    # endregion response methods

    # region static methods

    @classmethod
    def create(
        cls, ticket: str, namespace: str, platform_id: str, user_id: str, **kwargs
    ) -> PlatformLink:
        instance = cls()
        instance.ticket = ticket
        instance.namespace = namespace
        instance.platform_id = platform_id
        instance.user_id = user_id
        if x_flight_id := kwargs.get("x_flight_id", None):
            instance.x_flight_id = x_flight_id
        return instance

    @classmethod
    def create_from_dict(cls, dict_: dict, include_empty: bool = False) -> PlatformLink:
        instance = cls()
        if "ticket" in dict_ and dict_["ticket"] is not None:
            instance.ticket = str(dict_["ticket"])
        elif include_empty:
            instance.ticket = ""
        if "namespace" in dict_ and dict_["namespace"] is not None:
            instance.namespace = str(dict_["namespace"])
        elif include_empty:
            instance.namespace = ""
        if "platformId" in dict_ and dict_["platformId"] is not None:
            instance.platform_id = str(dict_["platformId"])
        elif include_empty:
            instance.platform_id = ""
        if "userId" in dict_ and dict_["userId"] is not None:
            instance.user_id = str(dict_["userId"])
        elif include_empty:
            instance.user_id = ""
        return instance

    @staticmethod
    def get_field_info() -> Dict[str, str]:
        return {
            "ticket": "ticket",
            "namespace": "namespace",
            "platformId": "platform_id",
            "userId": "user_id",
        }

    @staticmethod
    def get_required_map() -> Dict[str, bool]:
        return {
            "ticket": True,
            "namespace": True,
            "platformId": True,
            "userId": True,
        }

    # endregion static methods



# Copyright (c) 2021 AccelByte Inc. All Rights Reserved.
# This is licensed software from AccelByte Inc, for limitations
# and restrictions contact your company contract manager.
#
# Code generated. DO NOT EDIT!

# template file: ags_py_codegen

# AccelByte Gaming Services Platform Service

# pylint: disable=duplicate-code
# pylint: disable=line-too-long
# pylint: disable=missing-function-docstring
# pylint: disable=missing-module-docstring
# pylint: disable=too-many-arguments
# pylint: disable=too-many-branches
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-lines
# pylint: disable=too-many-locals
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-statements
# pylint: disable=unused-import

from __future__ import annotations
from typing import Any, Dict, List, Optional, Tuple, Union

from ....core import Model
from ....core import StrEnum


class AppTypeEnum(StrEnum):
    DEMO = "DEMO"
    DLC = "DLC"
    GAME = "GAME"
    SOFTWARE = "SOFTWARE"


class ClazzEnum(StrEnum):
    APP = "APP"
    CODE = "CODE"
    ENTITLEMENT = "ENTITLEMENT"
    LOOTBOX = "LOOTBOX"
    MEDIA = "MEDIA"
    OPTIONBOX = "OPTIONBOX"
    SUBSCRIPTION = "SUBSCRIPTION"


class OriginEnum(StrEnum):
    EPIC = "Epic"
    GOOGLEPLAY = "GooglePlay"
    IOS = "IOS"
    NINTENDO = "Nintendo"
    OCULUS = "Oculus"
    OTHER = "Other"
    PLAYSTATION = "Playstation"
    STEAM = "Steam"
    SYSTEM = "System"
    TWITCH = "Twitch"
    XBOX = "Xbox"


class StatusEnum(StrEnum):
    ACTIVE = "ACTIVE"
    CONSUMED = "CONSUMED"
    INACTIVE = "INACTIVE"
    REVOKED = "REVOKED"
    SOLD = "SOLD"


class TypeEnum(StrEnum):
    CONSUMABLE = "CONSUMABLE"
    DURABLE = "DURABLE"


class EntitlementIfc(Model):
    """Entitlement ifc (EntitlementIfc)

    Properties:
        app_id: (appId) OPTIONAL str

        app_type: (appType) OPTIONAL Union[str, AppTypeEnum]

        clazz: (clazz) OPTIONAL Union[str, ClazzEnum]

        collection_id: (collectionId) OPTIONAL str

        created_at: (createdAt) OPTIONAL str

        end_date: (endDate) OPTIONAL str

        features: (features) OPTIONAL List[str]

        granted_code: (grantedCode) OPTIONAL str

        id_: (id) OPTIONAL str

        item_id: (itemId) OPTIONAL str

        item_namespace: (itemNamespace) OPTIONAL str

        name: (name) OPTIONAL str

        namespace: (namespace) OPTIONAL str

        no_origin: (noOrigin) OPTIONAL bool

        origin: (origin) OPTIONAL Union[str, OriginEnum]

        sku: (sku) OPTIONAL str

        start_date: (startDate) OPTIONAL str

        status: (status) OPTIONAL Union[str, StatusEnum]

        store_id: (storeId) OPTIONAL str

        type_: (type) OPTIONAL Union[str, TypeEnum]

        updated_at: (updatedAt) OPTIONAL str

        use_count: (useCount) OPTIONAL int

        user_id: (userId) OPTIONAL str
    """

    # region fields

    app_id: str  # OPTIONAL
    app_type: Union[str, AppTypeEnum]  # OPTIONAL
    clazz: Union[str, ClazzEnum]  # OPTIONAL
    collection_id: str  # OPTIONAL
    created_at: str  # OPTIONAL
    end_date: str  # OPTIONAL
    features: List[str]  # OPTIONAL
    granted_code: str  # OPTIONAL
    id_: str  # OPTIONAL
    item_id: str  # OPTIONAL
    item_namespace: str  # OPTIONAL
    name: str  # OPTIONAL
    namespace: str  # OPTIONAL
    no_origin: bool  # OPTIONAL
    origin: Union[str, OriginEnum]  # OPTIONAL
    sku: str  # OPTIONAL
    start_date: str  # OPTIONAL
    status: Union[str, StatusEnum]  # OPTIONAL
    store_id: str  # OPTIONAL
    type_: Union[str, TypeEnum]  # OPTIONAL
    updated_at: str  # OPTIONAL
    use_count: int  # OPTIONAL
    user_id: str  # OPTIONAL

    # endregion fields

    # region with_x methods

    def with_app_id(self, value: str) -> EntitlementIfc:
        self.app_id = value
        return self

    def with_app_type(self, value: Union[str, AppTypeEnum]) -> EntitlementIfc:
        self.app_type = value
        return self

    def with_clazz(self, value: Union[str, ClazzEnum]) -> EntitlementIfc:
        self.clazz = value
        return self

    def with_collection_id(self, value: str) -> EntitlementIfc:
        self.collection_id = value
        return self

    def with_created_at(self, value: str) -> EntitlementIfc:
        self.created_at = value
        return self

    def with_end_date(self, value: str) -> EntitlementIfc:
        self.end_date = value
        return self

    def with_features(self, value: List[str]) -> EntitlementIfc:
        self.features = value
        return self

    def with_granted_code(self, value: str) -> EntitlementIfc:
        self.granted_code = value
        return self

    def with_id(self, value: str) -> EntitlementIfc:
        self.id_ = value
        return self

    def with_item_id(self, value: str) -> EntitlementIfc:
        self.item_id = value
        return self

    def with_item_namespace(self, value: str) -> EntitlementIfc:
        self.item_namespace = value
        return self

    def with_name(self, value: str) -> EntitlementIfc:
        self.name = value
        return self

    def with_namespace(self, value: str) -> EntitlementIfc:
        self.namespace = value
        return self

    def with_no_origin(self, value: bool) -> EntitlementIfc:
        self.no_origin = value
        return self

    def with_origin(self, value: Union[str, OriginEnum]) -> EntitlementIfc:
        self.origin = value
        return self

    def with_sku(self, value: str) -> EntitlementIfc:
        self.sku = value
        return self

    def with_start_date(self, value: str) -> EntitlementIfc:
        self.start_date = value
        return self

    def with_status(self, value: Union[str, StatusEnum]) -> EntitlementIfc:
        self.status = value
        return self

    def with_store_id(self, value: str) -> EntitlementIfc:
        self.store_id = value
        return self

    def with_type(self, value: Union[str, TypeEnum]) -> EntitlementIfc:
        self.type_ = value
        return self

    def with_updated_at(self, value: str) -> EntitlementIfc:
        self.updated_at = value
        return self

    def with_use_count(self, value: int) -> EntitlementIfc:
        self.use_count = value
        return self

    def with_user_id(self, value: str) -> EntitlementIfc:
        self.user_id = value
        return self

    # endregion with_x methods

    # region to methods

    def to_dict(self, include_empty: bool = False) -> dict:
        result: dict = {}
        if hasattr(self, "app_id"):
            result["appId"] = str(self.app_id)
        elif include_empty:
            result["appId"] = ""
        if hasattr(self, "app_type"):
            result["appType"] = str(self.app_type)
        elif include_empty:
            result["appType"] = Union[str, AppTypeEnum]()
        if hasattr(self, "clazz"):
            result["clazz"] = str(self.clazz)
        elif include_empty:
            result["clazz"] = Union[str, ClazzEnum]()
        if hasattr(self, "collection_id"):
            result["collectionId"] = str(self.collection_id)
        elif include_empty:
            result["collectionId"] = ""
        if hasattr(self, "created_at"):
            result["createdAt"] = str(self.created_at)
        elif include_empty:
            result["createdAt"] = ""
        if hasattr(self, "end_date"):
            result["endDate"] = str(self.end_date)
        elif include_empty:
            result["endDate"] = ""
        if hasattr(self, "features"):
            result["features"] = [str(i0) for i0 in self.features]
        elif include_empty:
            result["features"] = []
        if hasattr(self, "granted_code"):
            result["grantedCode"] = str(self.granted_code)
        elif include_empty:
            result["grantedCode"] = ""
        if hasattr(self, "id_"):
            result["id"] = str(self.id_)
        elif include_empty:
            result["id"] = ""
        if hasattr(self, "item_id"):
            result["itemId"] = str(self.item_id)
        elif include_empty:
            result["itemId"] = ""
        if hasattr(self, "item_namespace"):
            result["itemNamespace"] = str(self.item_namespace)
        elif include_empty:
            result["itemNamespace"] = ""
        if hasattr(self, "name"):
            result["name"] = str(self.name)
        elif include_empty:
            result["name"] = ""
        if hasattr(self, "namespace"):
            result["namespace"] = str(self.namespace)
        elif include_empty:
            result["namespace"] = ""
        if hasattr(self, "no_origin"):
            result["noOrigin"] = bool(self.no_origin)
        elif include_empty:
            result["noOrigin"] = False
        if hasattr(self, "origin"):
            result["origin"] = str(self.origin)
        elif include_empty:
            result["origin"] = Union[str, OriginEnum]()
        if hasattr(self, "sku"):
            result["sku"] = str(self.sku)
        elif include_empty:
            result["sku"] = ""
        if hasattr(self, "start_date"):
            result["startDate"] = str(self.start_date)
        elif include_empty:
            result["startDate"] = ""
        if hasattr(self, "status"):
            result["status"] = str(self.status)
        elif include_empty:
            result["status"] = Union[str, StatusEnum]()
        if hasattr(self, "store_id"):
            result["storeId"] = str(self.store_id)
        elif include_empty:
            result["storeId"] = ""
        if hasattr(self, "type_"):
            result["type"] = str(self.type_)
        elif include_empty:
            result["type"] = Union[str, TypeEnum]()
        if hasattr(self, "updated_at"):
            result["updatedAt"] = str(self.updated_at)
        elif include_empty:
            result["updatedAt"] = ""
        if hasattr(self, "use_count"):
            result["useCount"] = int(self.use_count)
        elif include_empty:
            result["useCount"] = 0
        if hasattr(self, "user_id"):
            result["userId"] = str(self.user_id)
        elif include_empty:
            result["userId"] = ""
        return result

    # endregion to methods

    # region static methods

    @classmethod
    def create(
        cls,
        app_id: Optional[str] = None,
        app_type: Optional[Union[str, AppTypeEnum]] = None,
        clazz: Optional[Union[str, ClazzEnum]] = None,
        collection_id: Optional[str] = None,
        created_at: Optional[str] = None,
        end_date: Optional[str] = None,
        features: Optional[List[str]] = None,
        granted_code: Optional[str] = None,
        id_: Optional[str] = None,
        item_id: Optional[str] = None,
        item_namespace: Optional[str] = None,
        name: Optional[str] = None,
        namespace: Optional[str] = None,
        no_origin: Optional[bool] = None,
        origin: Optional[Union[str, OriginEnum]] = None,
        sku: Optional[str] = None,
        start_date: Optional[str] = None,
        status: Optional[Union[str, StatusEnum]] = None,
        store_id: Optional[str] = None,
        type_: Optional[Union[str, TypeEnum]] = None,
        updated_at: Optional[str] = None,
        use_count: Optional[int] = None,
        user_id: Optional[str] = None,
        **kwargs,
    ) -> EntitlementIfc:
        instance = cls()
        if app_id is not None:
            instance.app_id = app_id
        if app_type is not None:
            instance.app_type = app_type
        if clazz is not None:
            instance.clazz = clazz
        if collection_id is not None:
            instance.collection_id = collection_id
        if created_at is not None:
            instance.created_at = created_at
        if end_date is not None:
            instance.end_date = end_date
        if features is not None:
            instance.features = features
        if granted_code is not None:
            instance.granted_code = granted_code
        if id_ is not None:
            instance.id_ = id_
        if item_id is not None:
            instance.item_id = item_id
        if item_namespace is not None:
            instance.item_namespace = item_namespace
        if name is not None:
            instance.name = name
        if namespace is not None:
            instance.namespace = namespace
        if no_origin is not None:
            instance.no_origin = no_origin
        if origin is not None:
            instance.origin = origin
        if sku is not None:
            instance.sku = sku
        if start_date is not None:
            instance.start_date = start_date
        if status is not None:
            instance.status = status
        if store_id is not None:
            instance.store_id = store_id
        if type_ is not None:
            instance.type_ = type_
        if updated_at is not None:
            instance.updated_at = updated_at
        if use_count is not None:
            instance.use_count = use_count
        if user_id is not None:
            instance.user_id = user_id
        return instance

    @classmethod
    def create_from_dict(
        cls, dict_: dict, include_empty: bool = False
    ) -> EntitlementIfc:
        instance = cls()
        if not dict_:
            return instance
        if "appId" in dict_ and dict_["appId"] is not None:
            instance.app_id = str(dict_["appId"])
        elif include_empty:
            instance.app_id = ""
        if "appType" in dict_ and dict_["appType"] is not None:
            instance.app_type = str(dict_["appType"])
        elif include_empty:
            instance.app_type = Union[str, AppTypeEnum]()
        if "clazz" in dict_ and dict_["clazz"] is not None:
            instance.clazz = str(dict_["clazz"])
        elif include_empty:
            instance.clazz = Union[str, ClazzEnum]()
        if "collectionId" in dict_ and dict_["collectionId"] is not None:
            instance.collection_id = str(dict_["collectionId"])
        elif include_empty:
            instance.collection_id = ""
        if "createdAt" in dict_ and dict_["createdAt"] is not None:
            instance.created_at = str(dict_["createdAt"])
        elif include_empty:
            instance.created_at = ""
        if "endDate" in dict_ and dict_["endDate"] is not None:
            instance.end_date = str(dict_["endDate"])
        elif include_empty:
            instance.end_date = ""
        if "features" in dict_ and dict_["features"] is not None:
            instance.features = [str(i0) for i0 in dict_["features"]]
        elif include_empty:
            instance.features = []
        if "grantedCode" in dict_ and dict_["grantedCode"] is not None:
            instance.granted_code = str(dict_["grantedCode"])
        elif include_empty:
            instance.granted_code = ""
        if "id" in dict_ and dict_["id"] is not None:
            instance.id_ = str(dict_["id"])
        elif include_empty:
            instance.id_ = ""
        if "itemId" in dict_ and dict_["itemId"] is not None:
            instance.item_id = str(dict_["itemId"])
        elif include_empty:
            instance.item_id = ""
        if "itemNamespace" in dict_ and dict_["itemNamespace"] is not None:
            instance.item_namespace = str(dict_["itemNamespace"])
        elif include_empty:
            instance.item_namespace = ""
        if "name" in dict_ and dict_["name"] is not None:
            instance.name = str(dict_["name"])
        elif include_empty:
            instance.name = ""
        if "namespace" in dict_ and dict_["namespace"] is not None:
            instance.namespace = str(dict_["namespace"])
        elif include_empty:
            instance.namespace = ""
        if "noOrigin" in dict_ and dict_["noOrigin"] is not None:
            instance.no_origin = bool(dict_["noOrigin"])
        elif include_empty:
            instance.no_origin = False
        if "origin" in dict_ and dict_["origin"] is not None:
            instance.origin = str(dict_["origin"])
        elif include_empty:
            instance.origin = Union[str, OriginEnum]()
        if "sku" in dict_ and dict_["sku"] is not None:
            instance.sku = str(dict_["sku"])
        elif include_empty:
            instance.sku = ""
        if "startDate" in dict_ and dict_["startDate"] is not None:
            instance.start_date = str(dict_["startDate"])
        elif include_empty:
            instance.start_date = ""
        if "status" in dict_ and dict_["status"] is not None:
            instance.status = str(dict_["status"])
        elif include_empty:
            instance.status = Union[str, StatusEnum]()
        if "storeId" in dict_ and dict_["storeId"] is not None:
            instance.store_id = str(dict_["storeId"])
        elif include_empty:
            instance.store_id = ""
        if "type" in dict_ and dict_["type"] is not None:
            instance.type_ = str(dict_["type"])
        elif include_empty:
            instance.type_ = Union[str, TypeEnum]()
        if "updatedAt" in dict_ and dict_["updatedAt"] is not None:
            instance.updated_at = str(dict_["updatedAt"])
        elif include_empty:
            instance.updated_at = ""
        if "useCount" in dict_ and dict_["useCount"] is not None:
            instance.use_count = int(dict_["useCount"])
        elif include_empty:
            instance.use_count = 0
        if "userId" in dict_ and dict_["userId"] is not None:
            instance.user_id = str(dict_["userId"])
        elif include_empty:
            instance.user_id = ""
        return instance

    @classmethod
    def create_many_from_dict(
        cls, dict_: dict, include_empty: bool = False
    ) -> Dict[str, EntitlementIfc]:
        return (
            {k: cls.create_from_dict(v, include_empty=include_empty) for k, v in dict_}
            if dict_
            else {}
        )

    @classmethod
    def create_many_from_list(
        cls, list_: list, include_empty: bool = False
    ) -> List[EntitlementIfc]:
        return (
            [cls.create_from_dict(i, include_empty=include_empty) for i in list_]
            if list_
            else []
        )

    @classmethod
    def create_from_any(
        cls, any_: any, include_empty: bool = False, many: bool = False
    ) -> Union[EntitlementIfc, List[EntitlementIfc], Dict[Any, EntitlementIfc]]:
        if many:
            if isinstance(any_, dict):
                return cls.create_many_from_dict(any_, include_empty=include_empty)
            elif isinstance(any_, list):
                return cls.create_many_from_list(any_, include_empty=include_empty)
            else:
                raise ValueError()
        else:
            return cls.create_from_dict(any_, include_empty=include_empty)

    @staticmethod
    def get_field_info() -> Dict[str, str]:
        return {
            "appId": "app_id",
            "appType": "app_type",
            "clazz": "clazz",
            "collectionId": "collection_id",
            "createdAt": "created_at",
            "endDate": "end_date",
            "features": "features",
            "grantedCode": "granted_code",
            "id": "id_",
            "itemId": "item_id",
            "itemNamespace": "item_namespace",
            "name": "name",
            "namespace": "namespace",
            "noOrigin": "no_origin",
            "origin": "origin",
            "sku": "sku",
            "startDate": "start_date",
            "status": "status",
            "storeId": "store_id",
            "type": "type_",
            "updatedAt": "updated_at",
            "useCount": "use_count",
            "userId": "user_id",
        }

    @staticmethod
    def get_required_map() -> Dict[str, bool]:
        return {
            "appId": False,
            "appType": False,
            "clazz": False,
            "collectionId": False,
            "createdAt": False,
            "endDate": False,
            "features": False,
            "grantedCode": False,
            "id": False,
            "itemId": False,
            "itemNamespace": False,
            "name": False,
            "namespace": False,
            "noOrigin": False,
            "origin": False,
            "sku": False,
            "startDate": False,
            "status": False,
            "storeId": False,
            "type": False,
            "updatedAt": False,
            "useCount": False,
            "userId": False,
        }

    @staticmethod
    def get_enum_map() -> Dict[str, List[Any]]:
        return {
            "appType": ["DEMO", "DLC", "GAME", "SOFTWARE"],
            "clazz": [
                "APP",
                "CODE",
                "ENTITLEMENT",
                "LOOTBOX",
                "MEDIA",
                "OPTIONBOX",
                "SUBSCRIPTION",
            ],
            "origin": [
                "Epic",
                "GooglePlay",
                "IOS",
                "Nintendo",
                "Oculus",
                "Other",
                "Playstation",
                "Steam",
                "System",
                "Twitch",
                "Xbox",
            ],
            "status": ["ACTIVE", "CONSUMED", "INACTIVE", "REVOKED", "SOLD"],
            "type": ["CONSUMABLE", "DURABLE"],
        }

    # endregion static methods
